"""
Rowan v2 API: pKa Workflow
Predict acid dissociation constants for ionizable groups in molecules.
"""

from typing import List, Dict, Any, Annotated
import rowan
import json
import stjames

def submit_pka_workflow(
    initial_molecule: Annotated[str, "SMILES string of the molecule to calculate pKa"],
    pka_range: Annotated[List[float], "pKa range [min, max] to search (e.g., [2, 12])"] = [2, 12],
    deprotonate_elements: Annotated[str, "Comma-separated elements for deprotonation (e.g., 'N,O,S'). Empty for auto-detect"] = "",
    protonate_elements: Annotated[str, "Comma-separated elements for protonation (e.g., 'N,O'). Empty for auto-detect"] = "",
    mode: Annotated[str, "Calculation mode: 'rapid', 'careful', or 'meticulous'"] = "careful",
    name: Annotated[str, "Workflow name for identification and tracking"] = "pKa Workflow",
    folder_uuid: Annotated[str, "UUID of folder to organize this workflow. Empty string uses default folder"] = "",
    max_credits: Annotated[int, "Maximum credits to spend on this calculation. 0 for no limit"] = 0
):
    """Submit a pKa prediction workflow using Rowan v2 API.
    
    Args:
        initial_molecule: The molecule to calculate the pKa of. SMILES string.
        pka_range: pKa range [min, max] to search, e.g., [2, 12]
        deprotonate_elements: Atomic numbers to consider for deprotonation, e.g., "[7, 8, 16]" for N, O, S. Empty string uses defaults.
        protonate_elements: Atomic numbers to consider for protonation, e.g., "[7, 8]" for N, O. Empty string uses defaults.
        mode: Calculation mode: 'rapid' (fast), 'careful' (balanced), or 'meticulous' (thorough)
        name: Workflow name for identification and tracking
        folder_uuid: UUID of folder to organize this workflow. Empty string uses default folder.
        max_credits: Maximum credits to spend on this calculation. 0 for no limit.
    
    Returns:
        Workflow object representing the submitted workflow
        
    Examples:
        # Phenol pKa
        result = submit_pka_workflow(
            initial_molecule="Oc1ccccc1",
            name="pKa phenol",
            deprotonate_elements="[8]"  # Only consider oxygen
        )
    """
    
    # Handle JSON string inputs for element lists
    parsed_deprotonate_elements = None
    if deprotonate_elements:
        try:
            parsed_deprotonate_elements = json.loads(deprotonate_elements)
        except json.JSONDecodeError:
            pass  # Keep as None if not valid JSON
            
    parsed_protonate_elements = None
    if protonate_elements:
        try:
            parsed_protonate_elements = json.loads(protonate_elements)
        except json.JSONDecodeError:
            pass  # Keep as None if not valid JSON

    # Convert List[float] to Tuple[float, float] for Rowan SDK compatibility
    pka_range_tuple = tuple(pka_range) if len(pka_range) == 2 else (pka_range[0], pka_range[1] if len(pka_range) > 1 else pka_range[0])
    
    return rowan.submit_pka_workflow(
        initial_molecule=stjames.Molecule.from_smiles(initial_molecule),
        pka_range=pka_range_tuple,
        deprotonate_elements=parsed_deprotonate_elements,
        protonate_elements=parsed_protonate_elements,
        mode=mode,
        name=name,
        folder_uuid=folder_uuid if folder_uuid else None,
        max_credits=max_credits if max_credits > 0 else None
    )


"""
Rowan spin states calculation function for MCP tool integration.
Follows the exact workflow from: https://github.com/rowansci/stjames-public/blob/master/stjames/workflows/spin_states.py
Implements SpinStatesWorkflow(MoleculeWorkflow) pattern from stjames.
"""

from typing import Optional, Union, List, Any
import logging
import rowan
from .molecular_converter import convert_to_smiles

logger = logging.getLogger(__name__)

# Mock stjames workflow classes to match their structure
class Mode:
    """Mock Mode enum from stjames."""
    AUTO = "auto"
    RAPID = "rapid" 
    CAREFUL = "careful"
    METICULOUS = "meticulous"
    RECKLESS = "reckless"
    MANUAL = "manual"

class Molecule:
    """Mock Molecule class from stjames."""
    def __init__(self, smiles: str, charge: int = 0, multiplicity: int = 1):
        self.smiles = smiles
        self.charge = charge
        self.multiplicity = multiplicity
    
    def __str__(self):
        return self.smiles

class MoleculeWorkflow:
    """
    Mock MoleculeWorkflow base class from stjames.
    Base class for Workflows that operate on a single molecule.
    """
    def __init__(self, initial_molecule: Union[str, Molecule], mode: str = "auto"):
        if isinstance(initial_molecule, str):
            self.initial_molecule = Molecule(initial_molecule)
        else:
            self.initial_molecule = initial_molecule
        
        # Set mode to RAPID if AUTO is selected (stjames behavior)
        if mode.lower() == "auto":
            self.mode = "rapid"
        else:
            self.mode = mode.lower()
    
    def __repr__(self):
        return f"<{type(self).__name__} {self.mode.upper()}>"

class SpinStatesWorkflow(MoleculeWorkflow):
    """
    Mock SpinStatesWorkflow from stjames that inherits from MoleculeWorkflow.
    Workflow for computing spin states of molecules.
    """
    def __init__(self, initial_molecule: Union[str, Molecule], states: List[int], 
                 mode: str = "auto", solvent: Optional[str] = None, 
                 xtb_preopt: bool = False, constraints: Optional[List] = None,
                 transition_state: bool = False, frequencies: bool = False):
        super().__init__(initial_molecule, mode)
        self.states = states
        self.solvent = solvent
        self.xtb_preopt = xtb_preopt
        self.constraints = constraints or []
        self.transition_state = transition_state
        self.frequencies = frequencies
        
        # Validate states (stjames validation logic)
        self._validate_states()
    
    def _validate_states(self):
        """Confirm that all spin states are valid (from stjames)."""
        if not self.states:
            raise ValueError("Expected at least one spin state.")
        
        # Check multiplicities are consistent (all odd or all even)
        if any((self.states[0] - mult) % 2 for mult in self.states):
            raise ValueError(f"Inconsistent multiplicities found: {self.states}")
    
    def __repr__(self):
        if self.mode != "manual":
            return f"<{type(self).__name__} {self.states} {self.mode.upper()}>"
        return f"<{type(self).__name__} {self.states} {self.mode.upper()}>"
    
    def __len__(self):
        return len(self.states)


# Removed old hardcoded conversion function - now using dynamic molecular_converter.py


def _generate_3d_coordinates_for_coordination_complex(smiles: str, rdkit_mol):
    """Generate clean 3D coordinates for coordination complexes."""
    try:
        import numpy as np
        
        # Get number of atoms
        num_atoms = rdkit_mol.GetNumAtoms()
        
        # Find metal center and ligands
        metal_idx = None
        ligand_indices = []
        
        for i, atom in enumerate(rdkit_mol.GetAtoms()):
            atomic_num = atom.GetAtomicNum()
            if atomic_num in [25, 26, 27, 28, 29, 24, 46]:  # Transition metals
                metal_idx = i
            else:
                ligand_indices.append(i)
        
        if metal_idx is None:
            return None
        
        # Initialize coordinates
        coords = [[0.0, 0.0, 0.0] for _ in range(num_atoms)]
        
        # Place metal at origin
        coords[metal_idx] = [0.0, 0.0, 0.0]
        
        # Set appropriate bond length for M-Cl bonds
        bond_length = 2.4  # Standard M-Cl bond length in Å
        
        # Generate geometry based on coordination number
        num_ligands = len(ligand_indices)
        
        if num_ligands == 4:
            # Square planar for Pd(II) complexes
            if any(atom.GetAtomicNum() == 46 for atom in rdkit_mol.GetAtoms()):
                positions = np.array([
                    [bond_length, 0, 0],
                    [-bond_length, 0, 0],
                    [0, bond_length, 0],
                    [0, -bond_length, 0]
                ])
            else:
                # Tetrahedral for other metals
                a = bond_length / np.sqrt(3)
                positions = np.array([
                    [a, a, a],
                    [a, -a, -a],
                    [-a, a, -a],
                    [-a, -a, a]
                ])
        
        elif num_ligands == 6:
            # Octahedral geometry
            positions = np.array([
                [bond_length, 0, 0],
                [-bond_length, 0, 0],
                [0, bond_length, 0],
                [0, -bond_length, 0],
                [0, 0, bond_length],
                [0, 0, -bond_length]
            ])
        
        else:
            # General spherical distribution
            angles = np.linspace(0, 2*np.pi, num_ligands, endpoint=False)
            positions = []
            for angle in angles:
                x = bond_length * np.cos(angle)
                y = bond_length * np.sin(angle)
                z = 0.0
                positions.append([x, y, z])
            positions = np.array(positions)
        
        # Assign positions to ligands
        for i, ligand_idx in enumerate(ligand_indices):
            if i < len(positions):
                coords[ligand_idx] = positions[i].tolist()
        
        # Validate minimum distances
        min_distance = float('inf')
        for i in range(num_atoms):
            for j in range(i+1, num_atoms):
                dist = np.linalg.norm(np.array(coords[i]) - np.array(coords[j]))
                min_distance = min(min_distance, dist)
        
        if min_distance > 1.0:  # At least 1.0 Å separation
            return coords
        else:
            return None
            
    except Exception as e:
        logger.warning(f"Coordinate generation failed: {e}")
        return None

# Removed old complex coordinate generation functions - simplified to single clean function above

def _calculate_molecular_charge(smiles: str) -> int:
    """Calculate the total molecular charge from SMILES string."""
    total_charge = 0
    
    # Count negative ions
    if '[Cl-]' in smiles:
        total_charge -= smiles.count('[Cl-]')
    if '[F-]' in smiles:
        total_charge -= smiles.count('[F-]')
    if '[Br-]' in smiles:
        total_charge -= smiles.count('[Br-]')
    if '[I-]' in smiles:
        total_charge -= smiles.count('[I-]')
    
    # Count positive metal ions - common oxidation states
    positive_ions = {
        '[Pd+2]': 2, '[Pd+4]': 4,
        '[Mn+2]': 2, '[Mn+3]': 3, '[Mn+4]': 4, '[Mn+7]': 7,
        '[Fe+2]': 2, '[Fe+3]': 3,
        '[Co+2]': 2, '[Co+3]': 3,
        '[Ni+2]': 2, '[Ni+3]': 3,
        '[Cu+1]': 1, '[Cu+2]': 2,
        '[Cr+2]': 2, '[Cr+3]': 3, '[Cr+6]': 6,
        '[V+2]': 2, '[V+3]': 3, '[V+4]': 4, '[V+5]': 5,
        '[Ti+2]': 2, '[Ti+3]': 3, '[Ti+4]': 4,
        '[Zn+2]': 2
    }
    
    for ion, charge in positive_ions.items():
        if ion in smiles:
            total_charge += charge * smiles.count(ion)
    
    return total_charge


def rowan_spin_states(
    name: str,
    molecule: str,
    states: Optional[Union[str, List[int]]] = None,
    mode: str = "rapid",
    solvent: Optional[str] = None,
    xtb_preopt: bool = False,
    constraints: Optional[str] = None,
    transition_state: bool = False,
    frequencies: bool = False,
    folder_uuid: Optional[str] = None,
    blocking: bool = False,
    ping_interval: int = 5
) -> str:
    """Calculate electronic spin states for molecular systems.
    
    This tool computes and compares different spin multiplicities for a molecule,
    helping determine the ground state electronic configuration and relative energies
    of different spin states. Essential for studying transition metals, radicals,
    and systems with unpaired electrons.
    
    Args:
        name: Name for the spin states calculation
        molecule: SMILES string of the molecule (e.g., "[Mn+3]", "O=O")
        states: List of spin multiplicities to calculate (e.g., [1, 3, 5] for singlet, triplet, quintet)
                Can be provided as comma-separated string "1,3,5" or list [1, 3, 5]
        mode: Calculation mode ("rapid", "careful", "meticulous", "auto", "reckless", "manual")
        solvent: Solvent for implicit solvation (e.g., "water", "hexane", "acetonitrile")
        xtb_preopt: Whether to perform xTB pre-optimization before DFT
        constraints: Geometric constraints during optimization (advanced feature)
        transition_state: Whether to optimize for transition state instead of minimum
        frequencies: Whether to calculate vibrational frequencies
        folder_uuid: Optional folder UUID to organize the calculation
        blocking: Whether to wait for completion (default: False to avoid timeouts)
        ping_interval: Interval in seconds to check calculation status
        
    Returns:
        JSON string with workflow UUID and status (non-blocking) or full results (blocking)
        
    Examples:
        # Manganese atom (stjames-public test example)
        rowan_spin_states(
            name="manganese_atom_stjames",
            molecule="[Mn]",  # Neutral Mn atom
            states=[2, 4, 6],  # doublet, quartet, sextet
            mode="rapid",
            xtb_preopt=True
        )
        
        # Iron atom (stjames-public test example)
        rowan_spin_states(
            name="iron_atom_stjames",
            molecule="[Fe]",  # Neutral Fe atom
            states=[1, 3, 5],  # singlet, triplet, quintet
            mode="careful",
            frequencies=True,
            xtb_preopt=True
        )
        
        # Manganese hexachloride complex (corrected format)
        rowan_spin_states(
            name="mn_hexachloride",
            molecule="[Cl-].[Cl-].[Cl-].[Cl-].[Cl-].[Cl-].[Mn+2]",
            states=[2, 4, 6],
            mode="rapid",
            solvent="water"
        )
    """
    
    # Convert input to SMILES format (handles XYZ coordinates, molecular formulas, etc.)
    original_input = molecule
    molecule = convert_to_smiles(molecule)
    
    # Log the conversion if it happened
    if molecule != original_input:
        logger.info(f"Converted input '{original_input}' to SMILES: '{molecule}'")
    
    # Check for unsupported formats
    if molecule.startswith("UNSUPPORTED_"):
        error_response = {
            "success": False,
            "error": f"Unsupported molecular input format: {original_input}",
            "explanation": {
                "detected_format": "XYZ coordinates or complex molecular structure",
                "supported_formats": [
                    "SMILES strings: '[Mn]', '[Cl-].[Mn+2]'",
                    "Simple formulas: 'Mn(Cl)6', 'Fe(Cl)6'",
                    "Common names: 'water', 'methane'"
                ]
            },
            "suggestion": "Please provide the molecule in SMILES format or a supported formula",
            "name": name
        }
        return str(error_response)
    
    # First, validate that molecule is actually a molecule (SMILES) and not a PDB file
    if molecule.lower().endswith('.pdb') or 'pdb' in molecule.lower():
        error_response = {
            "success": False,
            "error": f"Invalid input: '{molecule}' appears to be a PDB file",
            "correct_usage": {
                "purpose": "Spin states calculation requires a SMILES molecule string, not a PDB file",
                "explanation": "This tool calculates electronic spin states for individual molecules",
                "what_you_provided": "PDB file (protein structure)",
                "what_is_needed": "SMILES string representing a small molecule"
            },
            "examples": {
                "transition_metals": {
                    "manganese_atom": "[Mn] (neutral manganese atom)",
                    "iron_atom": "[Fe] (neutral iron atom)", 
                    "manganese_complex": "[Cl-].[Cl-].[Cl-].[Cl-].[Cl-].[Cl-].[Mn+2] (MnCl6 complex)"
                },
                "organic_molecules": {
                    "water": "O",
                    "methane": "C", 
                    "benzene": "c1ccccc1",
                    "radical": "c1ccc(cc1)[CH2] (benzyl radical)"
                }
            },
            "stjames_test_examples": {
                "manganese": "molecule='[Mn]', states=[2,4,6], mode='rapid'",
                "iron": "molecule='[Fe]', states=[1,3,5], mode='careful'"
            },
            "suggestion": "Use a different Rowan tool for protein analysis, or provide a SMILES string for molecular spin states",
            "name": name,
            "invalid_input": molecule
        }
        return str(error_response)
    
    # Check if molecule looks like it needs charge/oxidation state specification
    needs_charge_info = False
    molecular_formula = None
    
    # Detect if molecule needs proper SMILES formatting
    # Check for invalid SMILES patterns like [MnCl6]+4 or simple formulas like Mn(Cl)6
    invalid_patterns = [
        # Pattern like [MnCl6]+4 where charge is outside brackets
        (']' in molecule and ('+' in molecule[molecule.rfind(']'):] or '-' in molecule[molecule.rfind(']'):])),
        # Simple molecular formula without proper charge specification
        (not any(char in molecule for char in ['[', ']', '+', '-']) and 
         any(metal in molecule.upper() for metal in ['MN', 'FE', 'CO', 'NI', 'CU', 'CR', 'V', 'TI', 'PD'])),
        # Invalid complex notation like [MnCl6] without proper ion separation
        ('[' in molecule and ']' in molecule and 'Cl' in molecule.upper() and 
         not ('.' in molecule and '[Cl-]' in molecule))
    ]
    
    if any(invalid_patterns) and any(metal in molecule.upper() for metal in ['MN', 'FE', 'CO', 'NI', 'CU', 'CR', 'V', 'TI', 'PD']):
        needs_charge_info = True
        molecular_formula = molecule
    
    # Ensure we have states - provide intelligent defaults based on electron parity rules
    if states is None:
        # Auto-assign reasonable default states based on the molecule and electron count
        if any(metal in molecule.upper() for metal in ['MN', 'FE', 'CO', 'NI', 'CU', 'CR', 'V', 'TI', 'PD']):
            # For transition metals, assign states based on electron parity rules
            # Even electrons (Fe=26, Ni=28, Cr=24, Ti=22, V=23, Zn=30) need ODD multiplicities
            # Odd electrons (Mn=25, Co=27, Cu=29) need EVEN multiplicities
            
            if any(metal in molecule.upper() for metal in ['FE', 'NI', 'CR', 'TI', 'ZN', 'PD']):
                # Even electron metals → odd multiplicities
                states_list = [1, 3, 5]  # Singlet, triplet, quintet
                logger.info(f"Auto-assigned odd multiplicities for even-electron metal: {states_list}")
            elif any(metal in molecule.upper() for metal in ['MN', 'CO', 'CU', 'V']):
                # Odd electron metals → even multiplicities  
                states_list = [2, 4, 6]  # Doublet, quartet, sextet
                logger.info(f"Auto-assigned even multiplicities for odd-electron metal: {states_list}")
            else:
                # Default fallback for other transition metals
                states_list = [1, 3, 5]  # Conservative choice
                logger.info(f"Auto-assigned default states for transition metal: {states_list}")
        else:
            # For organic molecules, use singlet/triplet (most have even electrons)
            states_list = [1, 3]  # Singlet, triplet
            logger.info(f"Auto-assigned default states for organic molecule: {states_list}")
    
    # Handle states input - convert to list of integers
    if states is not None and isinstance(states, str):
        # Handle special keywords
        if states.lower() in ['all', 'comprehensive']:
            states_list = [1, 2, 3, 4, 5, 6]  # All reasonable multiplicities
        elif states.lower() in ['common', 'typical']:
            states_list = [2, 4, 6]  # Common for transition metals
        else:
            try:
                # Handle comma-separated string
                states_list = [int(s.strip()) for s in states.split(',')]
            except ValueError as e:
                error_response = {
                    "success": False,
                    "error": f"Invalid states format: '{states}'. Expected comma-separated integers, list, or keywords.",
                    "valid_formats": [
                        "Comma-separated: '2,4,6' or '1,3,5'",
                        "List: [2, 4, 6] or [1, 3, 5]", 
                        "Keywords: 'all' or 'comprehensive' for [1,2,3,4,5,6]",
                        "Keywords: 'common' or 'typical' for [2,4,6]"
                    ],
                    "name": name,
                    "molecule": molecule
                }
                return str(error_response)
    elif states is not None and isinstance(states, list):
        try:
            states_list = [int(s) for s in states]
        except (ValueError, TypeError) as e:
            error_response = {
                "success": False,
                "error": f"Invalid states list: {states}. All elements must be integers. Error: {e}",
                "name": name,
                "molecule": molecule
            }
            return str(error_response)
    elif states is not None:
        error_response = {
            "success": False,
            "error": f"Invalid states type: {type(states)}. Expected string or list.",
            "name": name,
            "molecule": molecule
        }
        return str(error_response)
    
    # Validate states list
    if not states_list:
        error_response = {
            "success": False,
            "error": "States list cannot be empty. Provide at least one spin multiplicity.",
            "name": name,
            "molecule": molecule
        }
        return str(error_response)
    
    # Validate states are positive
    for state in states_list:
        if state <= 0:
            error_response = {
                "success": False,
                "error": f"Invalid spin multiplicity: {state}. Must be positive integer.",
                "name": name,
                "molecule": molecule
            }
            return str(error_response)
    
    # Validate electron parity: even electrons need odd multiplicities, odd electrons need even multiplicities
    # This prevents impossible electronic configurations
    try:
        # Try to estimate electron count from SMILES (basic approach)
        # For complex molecules, we'll let Rowan handle the detailed validation
        # But we can catch obvious cases
        
        # Check if all states have the same parity (all odd or all even)
        first_state_parity = states_list[0] % 2
        mixed_parity = any((state % 2) != first_state_parity for state in states_list)
        
        if mixed_parity:
            error_response = {
                "success": False,
                "error": f"Inconsistent spin state parities: {states_list}. All multiplicities must be either odd (1,3,5...) or even (2,4,6...).",
                "explanation": {
                    "electron_parity_rule": "Molecules with even electrons need odd multiplicities; molecules with odd electrons need even multiplicities",
                    "your_states": states_list,
                    "fix_suggestions": {
                        "for_even_electrons": "Use odd multiplicities like [1,3,5]",
                        "for_odd_electrons": "Use even multiplicities like [2,4,6]"
                    }
                },
                "name": name,
                "molecule": molecule
            }
            return str(error_response)
    except Exception as e:
        # If electron counting fails, let Rowan handle the validation
        logger.warning(f"Could not validate electron parity: {e}")
        pass
    
    # Validate mode
    valid_modes = ["rapid", "careful", "meticulous", "auto", "reckless", "manual"]
    if mode.lower() not in valid_modes:
        error_response = {
            "success": False,
            "error": f"Invalid mode: {mode}. Valid modes: {valid_modes}",
            "name": name,
            "molecule": molecule
        }
        return str(error_response)
    
    # Create SpinStatesWorkflow instance to validate parameters (stjames pattern)
    try:
        spin_workflow = SpinStatesWorkflow(
            initial_molecule=molecule,
            states=states_list,
            mode=mode,
            solvent=solvent,
            xtb_preopt=xtb_preopt,
            constraints=constraints.split(',') if isinstance(constraints, str) else constraints,
            transition_state=transition_state,
            frequencies=frequencies
        )
        logger.info(f"SpinStatesWorkflow created: {spin_workflow}")
    except ValueError as e:
        error_response = {
            "success": False,
            "error": f"Invalid workflow parameters: {str(e)}",
            "name": name,
            "molecule": molecule,
            "states": states_list,
            "validation_error": str(e)
        }
        return str(error_response)
    
    logger.info(f"Starting spin states calculation: {name}")
    logger.info(f"Molecule SMILES: {molecule}")
    logger.info(f"Spin multiplicities: {states_list}")
    logger.info(f"Mode: {mode}")
    if solvent:
        logger.info(f"Solvent: {solvent}")
    if xtb_preopt:
        logger.info("xTB pre-optimization: enabled")
    if frequencies:
        logger.info("Frequency calculation: enabled")
    if transition_state:
        logger.info("Transition state optimization: enabled")
    
    # For spin states, we need to provide an initial multiplicity
    # Use the first state in the list as the starting multiplicity
    initial_multiplicity = states_list[0]
    
    # CRITICAL FIX: Ensure initial_multiplicity follows electron parity rules
    # Rowan uses initial_multiplicity as the starting point, so it must be valid
    logger.info(f"Using initial_multiplicity = {initial_multiplicity} for states {states_list}")
    
    # Prepare workflow parameters - rowan.compute() wants core params separate from workflow_data
    workflow_params = {
        # Core rowan.compute() parameters
        "molecule": molecule,  # SMILES string
        "workflow_type": "spin_states",
        "name": name,
        "blocking": blocking,
        "ping_interval": ping_interval,
        
        # CRITICAL: Try multiple parameter names for initial multiplicity
        "multiplicity": initial_multiplicity,  # Most likely parameter name
        "initial_multiplicity": initial_multiplicity,  # Alternative name
        "starting_multiplicity": initial_multiplicity,  # Another alternative
        "states": states_list,  # List of spin multiplicities to calculate
        "mode": mode.lower(),  # Calculation mode
    }
    
    # Add optional parameters only if provided
    if folder_uuid:
        workflow_params["folder_uuid"] = folder_uuid
    if solvent:
        workflow_params["solvent"] = solvent
    if xtb_preopt:
        workflow_params["xtb_preopt"] = xtb_preopt
    if transition_state:
        workflow_params["transition_state"] = transition_state
    if frequencies:
        workflow_params["frequencies"] = frequencies
    
    # Only add constraints if provided and ensure it's a list
    if constraints is not None:
        if isinstance(constraints, str):
            # Convert string to list (basic parsing)
            workflow_params["constraints"] = [constraints]
        elif isinstance(constraints, list):
            workflow_params["constraints"] = constraints
        else:
            workflow_params["constraints"] = [str(constraints)]
    
    logger.info("Submitting spin states calculation to Rowan")
    
    try:
        # CRITICAL FIX: Create Molecule object with correct multiplicity instead of string
        # This ensures Rowan uses the right starting multiplicity
        try:
            from stjames.molecule import Molecule as SjamesMolecule
            from rdkit import Chem
            
            # Create molecule with proper multiplicity by constructing atoms directly
            # Parse SMILES to get atomic numbers and coordinates
            rdkit_mol = Chem.MolFromSmiles(molecule)
            if rdkit_mol is None:
                raise ValueError(f"Invalid SMILES: {molecule}")
            
            # Get atoms from RDKit molecule and generate realistic 3D coordinates
            atoms = []
            from stjames.molecule import Atom
            
            # Generate realistic 3D coordinates using specialized coordination chemistry approach
            coords_3d = _generate_3d_coordinates_for_coordination_complex(molecule, rdkit_mol)
            
            for i, atom in enumerate(rdkit_mol.GetAtoms()):
                # Create Atom objects with atomic number and realistic 3D positions
                atoms.append(Atom(
                    atomic_number=atom.GetAtomicNum(), 
                    position=coords_3d[i]
                ))
            
            # Calculate total charge from SMILES string
            total_charge = _calculate_molecular_charge(molecule)
            
            logger.info(f"Calculated molecular charge: {total_charge} (from SMILES: {molecule})")
            
            # Create Molecule object directly with correct multiplicity and charge
            molecule_obj = SjamesMolecule(
                atoms=atoms,
                charge=total_charge,  # Calculate proper charge from SMILES
                multiplicity=initial_multiplicity,  # This is the key fix!
                smiles=molecule  # Keep original SMILES
            )
            workflow_params["molecule"] = molecule_obj
            logger.info(f"Created stjames.Molecule object with multiplicity={initial_multiplicity}")
        except (ImportError, Exception) as e:
            # If stjames not available, use string but log the limitation
            logger.warning(f"stjames.molecule.Molecule creation failed: {e}")
            logger.warning(f"Rowan may default to multiplicity=1 instead of {initial_multiplicity}")
        
        # Submit spin states calculation to Rowan
        result = rowan.compute(**workflow_params)
        
        # Format the response based on blocking mode
        if result:
            workflow_uuid = result.get("uuid")
            status = result.get("object_status", 0)
            
            if blocking and status == 2:  # Completed
                # Extract spin states results for completed blocking calls
                object_data = result.get("object_data", {})
                if "spin_states" in object_data:
                    spin_states_data = object_data["spin_states"]
                    
                    # Find ground state (lowest energy)
                    ground_state = None
                    min_energy = float('inf')
                    for state_data in spin_states_data:
                        energy = state_data.get("energy", float('inf'))
                        if energy < min_energy:
                            min_energy = energy
                            ground_state = state_data.get("multiplicity")
                    
                    response = {
                        "success": True,
                        "workflow_uuid": workflow_uuid,
                        "name": name,
                        "molecule": molecule,
                        "status": "completed",
                        "spin_states_results": {
                            "ground_state_multiplicity": ground_state,
                            "ground_state_energy": min_energy,
                            "states_calculated": len(spin_states_data),
                            "spin_states": spin_states_data
                        },
                        "calculation_details": {
                            "mode": mode,
                            "solvent": solvent,
                            "xtb_preopt": xtb_preopt,
                            "frequencies_calculated": frequencies,
                            "transition_state": transition_state
                        },
                        "runtime_seconds": result.get("elapsed", 0),
                        "credits_charged": result.get("credits_charged", 0)
                    }
                else:
                    response = {
                        "success": True,
                        "workflow_uuid": workflow_uuid,
                        "name": name,
                        "molecule": molecule,
                        "status": "completed", 
                        "message": "Spin states calculation completed successfully",
                        "runtime_seconds": result.get("elapsed", 0),
                        "credits_charged": result.get("credits_charged", 0)
                    }
            else:
                # Non-blocking or still running - return workflow info for tracking
                status_text = {0: "queued", 1: "running", 2: "completed", 3: "failed"}.get(status, "unknown")
                response = {
                    "success": True,
                    "tracking_id": workflow_uuid,  # Prominent tracking ID
                    "workflow_uuid": workflow_uuid,  # Keep for backward compatibility
                    "name": name,
                    "molecule": molecule,
                    "status": status_text,
                    "message": f"Spin states calculation submitted successfully! Use tracking_id to monitor progress.",
                    "calculation_details": {
                        "spin_multiplicities": states_list,
                        "mode": mode,
                        "solvent": solvent,
                        "xtb_preopt": xtb_preopt,
                        "frequencies": frequencies,
                        "transition_state": transition_state,
                        "blocking_mode": blocking
                    },
                    "progress_tracking": {
                        "tracking_id": workflow_uuid,
                        "check_status": f"rowan_workflow_management(action='status', workflow_uuid='{workflow_uuid}')",
                        "get_results": f"rowan_workflow_management(action='retrieve', workflow_uuid='{workflow_uuid}')"
                    }
                }
        else:
            response = {
                "success": False,
                "error": "No response received from Rowan API",
                "name": name,
                "molecule": molecule,
                "states": states_list
            }
            
        return str(response)
        
    except Exception as e:
        error_response = {
            "success": False,
            "error": f"Spin states calculation failed: {str(e)}",
            "name": name,
            "molecule": molecule,
            "states": states_list
        }
        logger.error(f"Spin states calculation failed: {str(e)}")
        return str(error_response)




"""
Main entry point for Rowan MCP Server when run as a module.

Usage:
    python -m rowan_mcp          # HTTP/SSE mode
    python -m rowan_mcp --help   # Show help
"""

if __name__ == "__main__":
    # HTTP transport only
    from .server import main
    main() 


"""
Rowan conformers function for conformational analysis.
"""

import os
import logging
import time
from typing import Optional

try:
    import rowan
except ImportError:
    rowan = None

# Setup logging
logger = logging.getLogger(__name__)



def log_rowan_api_call(workflow_type: str, **kwargs):
    """Log Rowan API calls and let Rowan handle its own errors."""
    
    # Simple logging for long-running calculations
    if workflow_type in ["multistage_opt", "conformer_search"]:
        blocking = kwargs.get('blocking', True)
        if blocking:
            logger.info(f" Starting {workflow_type.replace('_', ' ')}...")
        else:
            logger.info(f" Submitting {workflow_type.replace('_', ' ')} without waiting")
    
    # Let Rowan handle everything - no custom error handling
    return rowan.compute(workflow_type=workflow_type, **kwargs)

def rowan_conformers(
    name: str,
    molecule: str,
    max_conformers: int = 50,
    mode: str = "rapid",
    folder_uuid: Optional[str] = None,
    blocking: bool = True,
    ping_interval: int = 5
):
    """Generate and optimize molecular conformers using Rowan's conformer_search workflow. Valid modes are "reckless", "rapid", "careful", and "meticulous", and default to using SMILES strings for the "molecule" parameter.
    
    Args:
        name: Name for the calculation
        molecule: Molecule SMILES string or common name
        max_conformers: Maximum number of conformers to generate (default: 50)
        mode: Conformer search mode - "reckless", "rapid", "careful", "meticulous" (default: "rapid")
        folder_uuid: UUID of folder to organize calculation in
        blocking: Whether to wait for completion (default: True)
        ping_interval: How often to check status in seconds (default: 5)
    
    Returns:
        Conformer search results (actual results if blocking=True)
    """
    
    # Validate mode parameter
    valid_modes = ["reckless", "rapid", "careful", "meticulous"]
    if mode not in valid_modes:
        raise ValueError(
            f"Invalid mode '{mode}'. Valid modes are: {', '.join(valid_modes)}"
        )
    
    return log_rowan_api_call(
        workflow_type="conformer_search",
        name=name,
        molecule=molecule,
        mode=mode,
        max_conformers=max_conformers,
        folder_uuid=folder_uuid,
        blocking=blocking,
        ping_interval=ping_interval
    )

if __name__ == "__main__":
    pass 


"""
Rowan v2 API: Workflow Management Tools
MCP tools for interacting with Workflow objects returned by v2 API submission functions.
"""

from typing import Dict, Any, List, Annotated
import rowan


def workflow_get_status(
    workflow_uuid: Annotated[str, "UUID of the workflow to check status"]
) -> Dict[str, Any]:
    """Get the current status of a workflow.
    
    Args:
        workflow_uuid: UUID of the workflow to check status
    
    Returns:
        Dictionary with status information
    """
    workflow = rowan.retrieve_workflow(workflow_uuid)
    status = workflow.get_status()
    
    return {
        "uuid": workflow_uuid,
        "status": status,
        "name": workflow.name,
        "is_finished": workflow.is_finished()
    }


def workflow_wait_for_result(
    workflow_uuid: Annotated[str, "UUID of the workflow to wait for completion"],
    poll_interval: Annotated[int, "Seconds between status checks while waiting"] = 5
) -> Dict[str, Any]:
    """Wait for a workflow to complete and return the result.
    
    Args:
        workflow_uuid: UUID of the workflow to wait for completion
        poll_interval: Seconds between status checks while waiting
    
    Essential for chaining dependent workflows where subsequent calculations 
    require results from previous ones. Blocks execution until the workflow 
    completes, then returns the full results.
    
    Common use cases:
    - Conformer search → Redox potential for each conformer
    - Optimization → Frequency calculation on optimized geometry
    - Multiple sequential optimizations with different methods
    - Any workflow chain where results feed into next calculation
    
    Example workflow chain:
        1. Submit conformer search
        2. Wait for conformer search to complete (using this function)
        3. Extract conformer geometries from results
        4. Submit new workflows using those geometries
    
    Returns:
        Dictionary containing the completed workflow data including results
        needed for dependent workflows (e.g., conformer_uuids, optimized
        geometries, calculation_uuids)
    """
    workflow = rowan.retrieve_workflow(workflow_uuid)
    
    # Use the built-in wait_for_result method
    workflow.wait_for_result(poll_interval=poll_interval)
    
    # Return complete workflow data
    return {
        "uuid": workflow.uuid,
        "status": workflow.status,
        "name": workflow.name,
        "created_at": workflow.created_at,
        "updated_at": workflow.updated_at,
        "completed_at": workflow.completed_at,
        "parent_uuid": workflow.parent_uuid,
        "workflow_type": workflow.workflow_type,
        "data": workflow.data,
        "credits_charged": workflow.credits_charged,
        "elapsed": workflow.elapsed
    }


def workflow_stop(
    workflow_uuid: Annotated[str, "UUID of the running workflow to stop"]
) -> Dict[str, str]:
    """Stop a running workflow.
    
    Args:
        workflow_uuid: UUID of the running workflow to stop
    
    Returns:
        Dictionary with confirmation message
    """
    workflow = rowan.retrieve_workflow(workflow_uuid)
    workflow.stop()
    
    return {
        "message": f"Workflow {workflow_uuid} stop request submitted",
        "uuid": workflow_uuid
    }


def workflow_delete(
    workflow_uuid: Annotated[str, "UUID of the workflow to permanently delete"]
) -> Dict[str, str]:
    """Delete a workflow.
    
    Args:
        workflow_uuid: UUID of the workflow to permanently delete
    
    This permanently removes the workflow and its results from the database.
    
    Returns:
        Dictionary with confirmation message
    """
    workflow = rowan.retrieve_workflow(workflow_uuid)
    workflow.delete()
    
    return {
        "message": f"Workflow {workflow_uuid} deleted successfully",
        "uuid": workflow_uuid
    }


def workflow_fetch_latest(
    workflow_uuid: Annotated[str, "UUID of the workflow to fetch latest data"],
    in_place: Annotated[bool, "Whether to update the workflow object in place"] = False
) -> Dict[str, Any]:
    """Fetch the latest workflow data from the database.
    
    Args:
        workflow_uuid: UUID of the workflow to fetch latest data
        in_place: Whether to update the workflow object in place
    
    Updates the workflow object with the most recent status and results.
    
    Returns:
        Dictionary containing the updated workflow data including status and results
    """
    workflow = rowan.retrieve_workflow(workflow_uuid)
    
    # Fetch latest updates
    workflow.fetch_latest(in_place=in_place)
    
    # Return workflow data as dict
    return {
        "uuid": workflow.uuid,
        "status": workflow.status,
        "name": workflow.name,
        "created_at": workflow.created_at,
        "updated_at": workflow.updated_at,
        "started_at": workflow.started_at,
        "completed_at": workflow.completed_at,
        "parent_uuid": workflow.parent_uuid,
        "workflow_type": workflow.workflow_type,
        "data": workflow.data,
        "notes": workflow.notes,
        "starred": workflow.starred,
        "public": workflow.public,
        "credits_charged": workflow.credits_charged,
        "elapsed": workflow.elapsed
    }


def retrieve_workflow(
    uuid: Annotated[str, "UUID of the workflow to retrieve"]
) -> Dict[str, Any]:
    """Retrieve a workflow from the API.
    
    Args:
        uuid: UUID of the workflow to retrieve
    
    Returns:
        Dictionary containing the complete workflow data
        
    Raises:
        HTTPError: If the API request fails
    """
    workflow = rowan.retrieve_workflow(uuid)
    
    # Convert workflow object to dict
    return {
        "uuid": workflow.uuid,
        "status": workflow.status,
        "name": workflow.name,
        "created_at": workflow.created_at,
        "updated_at": workflow.updated_at,
        "started_at": workflow.started_at,
        "completed_at": workflow.completed_at,
        "parent_uuid": workflow.parent_uuid,
        "workflow_type": workflow.workflow_type,
        "data": workflow.data,
        "notes": workflow.notes,
        "starred": workflow.starred,
        "public": workflow.public,
        "email_when_complete": workflow.email_when_complete,
        "max_credits": workflow.max_credits,
        "elapsed": workflow.elapsed,
        "credits_charged": workflow.credits_charged
    }


def list_workflows(
    parent_uuid: Annotated[str, "UUID of parent folder to filter by. Empty string for all folders"] = "",
    name_contains: Annotated[str, "Substring to search for in workflow names. Empty string for all names"] = "",
    public: Annotated[str, "Filter by public status ('true'/'false'). Empty string for both"] = "",
    starred: Annotated[str, "Filter by starred status ('true'/'false'). Empty string for both"] = "",
    status: Annotated[str, "Filter by workflow status code. Empty string for all statuses"] = "",
    workflow_type: Annotated[str, "Filter by workflow type (e.g., 'conformer_search', 'pka'). Empty string for all types"] = "",
    page: Annotated[int, "Page number for pagination (0-indexed)"] = 0,
    size: Annotated[int, "Number of workflows per page"] = 10
):
    """List workflows subject to the specified criteria.
    
    Args:
        parent_uuid: UUID of parent folder to filter by. Empty string for all folders
        name_contains: Substring to search for in workflow names. Empty string for all names
        public: Filter by public status ("true"/"false"). Empty string for both
        starred: Filter by starred status ("true"/"false"). Empty string for both
        status: Filter by workflow status code. Empty string for all statuses
        workflow_type: Filter by workflow type (e.g., 'conformer_search', 'pka'). Empty string for all types
        page: Page number for pagination (0-indexed)
        size: Number of workflows per page
    
    Returns:
        List of workflow dictionaries that match the search criteria
        
    Raises:
        HTTPError: If the request to the API fails
    """
    # Use direct API call to avoid Workflow validation issues
    with rowan.api_client() as client:
        params = {
            "page": page,
            "size": size
        }
        
        # Add non-empty filters
        if parent_uuid:
            params["parent_uuid"] = parent_uuid
        if name_contains:
            params["name_contains"] = name_contains
        if public:
            params["public"] = public.lower() == "true"
        if starred:
            params["starred"] = starred.lower() == "true"
        if status:
            params["object_status"] = int(status)
        if workflow_type:
            params["object_type"] = workflow_type
        
        response = client.get("/workflow", params=params)
        response.raise_for_status()
        
        data = response.json()
        # Extract workflows from the paginated response
        return data.get("workflows", [])


def retrieve_calculation_molecules(
    uuid: Annotated[str, "UUID of the calculation to retrieve molecules from"]
) -> List[Dict[str, Any]]:
    """Retrieve a list of molecules from a calculation.
    
    Args:
        uuid: UUID of the calculation to retrieve molecules from
    
    Returns:
        List of dictionaries representing the molecules in the calculation
        
    Raises:
        HTTPError: If the API request fails
    """
    molecules = rowan.retrieve_calculation_molecules(uuid)
    
    # Convert molecules to list of dicts
    result = []
    for mol in molecules:
        mol_dict = {
            "smiles": mol.get("smiles"),
            "name": mol.get("name"),
            "charge": mol.get("charge"),
            "multiplicity": mol.get("multiplicity"),
            "energy": mol.get("energy"),
            "coordinates": mol.get("coordinates"),
            "properties": mol.get("properties", {})
        }
        # Remove None values
        mol_dict = {k: v for k, v in mol_dict.items() if v is not None}
        result.append(mol_dict)
    
    return result


def workflow_update(
    workflow_uuid: Annotated[str, "UUID of the workflow to update"],
    name: Annotated[str, "New name for the workflow. Empty string to keep current name"] = "",
    notes: Annotated[str, "New notes/description for the workflow. Empty string to keep current notes"] = "",
    starred: Annotated[str, "Set starred status ('true'/'false'). Empty string to keep current status"] = "",
    public: Annotated[str, "Set public visibility ('true'/'false'). Empty string to keep current status"] = ""
) -> Dict[str, Any]:
    """Update workflow details.
    
    Args:
        workflow_uuid: UUID of the workflow to update
        name: New name for the workflow. Empty string to keep current name
        notes: New notes/description for the workflow. Empty string to keep current notes
        starred: Set starred status ("true"/"false"). Empty string to keep current status
        public: Set public visibility ("true"/"false"). Empty string to keep current status
    
    Returns:
        Dictionary with updated workflow information
    """
    workflow = rowan.retrieve_workflow(workflow_uuid)
    
    # Parse string boolean inputs
    parsed_starred = None
    if starred:
        parsed_starred = starred.lower() == "true"
    
    parsed_public = None
    if public:
        parsed_public = public.lower() == "true"
    
    # Update the workflow
    workflow.update(
        name=name if name else None,
        notes=notes if notes else None,
        starred=parsed_starred,
        public=parsed_public
    )
    
    return {
        "uuid": workflow.uuid,
        "name": workflow.name,
        "notes": workflow.notes,
        "starred": workflow.starred,
        "public": workflow.public,
        "message": "Workflow updated successfully"
    }


def workflow_is_finished(
    workflow_uuid: Annotated[str, "UUID of the workflow to check completion status"]
) -> Dict[str, Any]:
    """Check if a workflow is finished.
    
    Args:
        workflow_uuid: UUID of the workflow to check completion status
    
    Returns:
        Dictionary with workflow completion status
    """
    workflow = rowan.retrieve_workflow(workflow_uuid)
    
    return {
        "uuid": workflow_uuid,
        "is_finished": workflow.is_finished(),
        "status": workflow.status,
        "name": workflow.name
    }


def workflow_delete_data(
    workflow_uuid: Annotated[str, "UUID of the workflow whose data to delete (keeps workflow record)"]
) -> Dict[str, str]:
    """Delete workflow data while keeping the workflow record.
    
    Args:
        workflow_uuid: UUID of the workflow whose data to delete (keeps workflow record)
    
    Returns:
        Dictionary with confirmation message
    """
    workflow = rowan.retrieve_workflow(workflow_uuid)
    workflow.delete_data()
    
    return {
        "message": f"Data for workflow {workflow_uuid} deleted successfully",
        "uuid": workflow_uuid
    }


import time

import stjames

import rowan

HARTREE_TO_KCALMOL = 627.5096

ligands = [
    "CCC(C)(C)NC1=NCC2(CCC(=O)C2C)N1",
    "CCC(C)CN=C1NCC2(CCCOC2)CN1",
    "CC(C)CCNC1=NCC2CC(COC2=N)O1",
    "CCC(CC)NC1=NCC2CC(CO)CC12",
    "CCC(C)CN=C1NC=C2CCC(O)CC2=N1",
    "CC(C#CC#N)C1=NC=C2C=CC(=O)CN12",
    "CCC(CCN)C1NCC2=C(CC)SC(O)=C12",
    "CCC(CC)N=C1NC=C2C=C(F)C=C(N)N12",
    "CC(C)(C)C(N)C1=NCC2=CC(I)=NC=C2O1",
    "CCC(CC)NC1=NCC2CC(N1)C(C)CN2",
    "CCCC(C)N=C1NCC2CCNC2(C)CO1",
    "CC(C)CC(N)C1NCC2=CC(N)=C(Br)N=C12",
    "CC(C)C(CN)C1=NC=C2C=CN(C)C2=C1Br",
    "CC(CC#C)NC1=NCC2=C(C)N(C)C=C2O1",
    "CCCC(C)(N)C1=NCC2=CC(N)=CC=C2O1",
    "CCCCC(N)C1=NC=C2C=CN(C)C(C)=C12",
    "CCC(CC)N=C1NCC2CC(N)CCCN12",
    "CC(C)C(C)(N)C1=NC=C2C=CNC(=N)C2=N1",
    "CC(C)(CC#N)C1=NCC2CCN=COC2O1",
    "C#CCC(C#N)C1=NC=C2C=CNN=C12",
    "CCCCCNC1=NCC2=C(C)NN=C2N1",
    "CCCCC(N)C1=NC=C2C(C)=NOC2=C1Br",
    "CCCCCNC1=NCC2(CC#N)OCC1O2",
    "CC(C)CCNC1=NC=C2CC(=N)OCCN12",
    "CCCC(C#N)C1N(C)C2=C(C=NS2)C1=O",
    "CCC(CC#N)C1=NCC2=C(CO1)OC(=N)S2",
    "CCC(CCN)C1=NCC2=C(CO1)SC=N2",
    "CCC(C(C)N)C1NCC2CCOC1C2(C)C",
    "CC(C)(C)CNC1=NCC2CCOC1(O2)C#C",
    "CCCCCN=C1N(C)C2CCOC2C1(C)C",
    "CC(C)CCN=C1NCC2(CCOC2O1)C#C",
    "CC(C)C(C#N)C1NCC2=CC(O)=C(Br)N=C12",
    "CCC(C(C)N)C1(N)CC2CC(OC)C1C2C",
    "CC(C)(C)CN=C1NC=C2CC(O)CC2O1",
    "CCC(C)(C#N)C1=NC=C2CC(=O)C(C=O)N12",
    "CCCC(C)NC1=NCC2(CCOC=N2)CC1",
    "CCCC(C)N=C1N=CC2=CC(=O)N=C2C1=N",
    "CC(C)C(C#N)C1=NC=C2C=C(O)OC2=C1N",
    "CCCC(CN=C1NCC2=C(C)SC=C12)C=O",
    "CC(C)(C)CN=C1NCC2=CNC3=C2C1=CN3",
    "CCCCC(N)C1=NC=C2CNCC2=C1I",
    "CC(CC#C)NC1=NCC2=C(NC(=C2)C#C)N1",
    "CCC#CCN=C1NCC2=CNC=C2N1",
    "CCCC(C#N)C1N(C)C2=C(NC=C2)OC1=O",
    "CC(C)(C)CN=C1NC=C2C=NC=C2SC1=N",
    "CC(C)C(C#N)C1NCC2=CN=C(C=C12)C#N",
    "CC(C)C(C#N)C1(N)CC2CN(C)CC2C1C",
    "CCC(C)(C)N=C1NCC2=C(NC(C)=C2)C1O",
    "CC(C)C(C)NC1=NCC2CNC(C)C(C2)C1",
    "CCC(C)CN=C1NCC2C(N)C(C)C(C)C12",
    "CCC(C)(C#N)C1=NC=C2CNC(C)C(C)N12",
    "CC(C)C(C#N)C1NCC2=C(N=C(C)N2)C1C",
    "CCC(C)(C)NC1=NCC2=C(N=C(C)N2)C1C",
    "CCC(C)(C)N=C1N=CC2=CN(C)N=C2N1C",
    "CC(C)CCN=C1NC=C2C=NC(=NN12)C#C",
    "CC(C)C(CN=C1NCC2(CN=CO2)O1)C#N",
    "CCC(CC)N=C1N(C)C2=C(NC(=O)S2)C1=O",
    "CCCC(C#N)C1NCC2=C(N=CS2)S1(=O)=O",
    "CCC(C)(C)NC1=NCC2=CNN=C2C(=N)N1",
    "CC(C)CCN=C1NC=C2C(N)=NC(C)=C2S1",
    "CCC(C)(C)NC1=NC=C2C(=N)N=CNC2=N1",
    "CCC(CC)N=C1N(C)C2=C(NN=N2)C1=NO",
    "CCC(C)(C)NC1=NCC2=CN=NN2CCC1",
    "CCC(C)CN=C1N=CC2=CNN=NC2=N1",
    "CCC(CCN)C1=NC=C2C(N)=NSC2=N1",
    "CCC(C)(CN)C1NCC2=C(N)OC=C2C1C",
    "CC(C)(CCN)C1=NC=C2C(=N)SN=C2C=N1",
    "CC(C)(C)CNC1=NC=C2C(=N)SN=C2O1",
    "CCC(C)C(N)C1=NCC2=C(O1)C=CC(=N)S2",
    "CCCCCN=C1NCC2=C(O1)N=C(S2)C#N",
    "CCC(C)(CNC1=NCC2=C(O1)SC=N2)C=O",
    "CC#CC(CN)C1=NCC2=C(O1)SN=C2C",
    "CCCC(C#N)C1NCC2COC1(C2)C(C)=O",
    "CC(C)(C)C(N)C1(N)CC2COC1C(C)(C)C2",
    "CCCC(C#N)C1NCC2COC1(C)CN2C",
    "CCC(C)(C)N=C1N(C)C2COC1(C)OC2C",
    "CC(C)CCN=C1NCC2=C(OC1=N)N=NN2",
    "CCC(CC#N)(C1NCC2=COC=C12)N(C)C",
    "CCCC(C)(N)C1NCC2=C(OC=C12)N(C)C",
    "CC(C)(C)CNC1=NCC2=C(OCC1)N=CS2",
    "CC#CCCN=C1NCC2=COC=C2C=C1",
    "CCCCCNC1=NC=C2C(OCC2(C)O)=C1",
    "CCC(C)(C#N)C1N(C)C2=COC=C2N=C1N",
    "CCC(C)(C)N=C1NCC2(C)OC(CC12)=NC",
    "CCC(C)(CN)C1NCC2C(O)CCC1C2C",
    "CCCC(C)NC1=NCC2COCC(C1)C2C",
    "CC(C)C(C)NC1=NCC2COCCC2(C)O1",
    "CC(C)CC(N)C1=NC=C2COC(C)C(C)N12",
    "CCC(C)(C)NC1=NCC2COC(C)(O1)C2N",
    "CCCC(C)(N)C1=NC=C2COCCOC2=N1",
    "CCCC(C)N=C1N(C)C2=COC(N)=C2C1=O",
    "CCCCCNC1=NCC2(C)OC=NCCC12",
    "CCC(C(C)N)C1=NC=C2COC(N)=NC2=C1",
    "CCCC(C)NC1=NCC2(C)OC(=O)OC2O1",
    "CCCCC(N)C1NCC2=CON=C2C1C#C",
    "CC(C)C(C)N=C1NCC2=C(O)N(C)C=C2O1",
    "CCCC(C)N=C1N=CC2=C(O)N(C)C=CN12",
    "CC(C)C(C#N)C1=NC=C2C(O)=NSC2=C1O",
    "CCC(CC)N=C1NCC2=C(O)SC(=N)N=C12",
    "CC(C)C(CN)C1NCC2=C(SC=C2C)C1C",
]


workflows = []
results = {}

docking_result_folder = rowan.create_folder(name="Docking results")


protein = rowan.create_protein_from_pdb_id("CDK2", "1HCK")

protein.sanitize()
time.sleep(60)
protein.refresh()

for ligand in ligands:
    workflow = rowan.submit_docking_workflow(
        protein.uuid,
        pocket=[[103.55, 100.59, 82.99], [27.76, 32.67, 48.79]],
        initial_molecule=stjames.Molecule.from_smiles(ligand),
        folder_uuid=docking_result_folder.uuid,
        name=f"Docking {ligand}",
    )
    workflows.append(workflow)


for workflow in workflows:
    workflow.wait_for_result()
    workflow.fetch_latest(in_place=True)

lowest_conformer_energy = 0
for workflow in workflows:
    conformers = workflow.data["conformers"]
    for conformer in conformers:
        energy = rowan.retrieve_calculation_molecules(
            conformer[0] if isinstance(conformer, list) else conformer
        )[0]["energy"]
        lowest_conformer_energy = min(lowest_conformer_energy, energy)

    sorted_scores = sorted(workflow.data["scores"], key=lambda x: float(x["score"]))
    for score in sorted_scores:
        pose_energy = rowan.retrieve_calculation_molecules(score["pose"])[0]["energy"]
        strain = (pose_energy - lowest_conformer_energy) * HARTREE_TO_KCALMOL
        if score["posebusters_valid"] and strain < 4:
            results[workflow.name] = score
            break

print(results)


"""
Rowan scan function for potential energy surface scans and IRC workflows.
"""

from typing import Optional, Union, List, Dict, Any
import rowan
import logging
import os

# Set up logger
logger = logging.getLogger(__name__)

# Get API key from environment
api_key = os.environ.get("ROWAN_API_KEY")
if api_key:
    rowan.api_key = api_key
else:
    logger.warning("ROWAN_API_KEY not found in environment")

# QC Constants needed for validation
QC_ENGINES = {
    "psi4": "Hartree–Fock and density-functional theory",
    "terachem": "Hartree–Fock and density-functional theory", 
    "pyscf": "Hartree–Fock and density-functional theory",
    "xtb": "Semiempirical calculations",
    "aimnet2": "Machine-learned interatomic potential calculations"
}

QC_METHODS = {
    # Hartree-Fock
    "hf": "Hartree-Fock (unrestricted for open-shell systems)",
    
    # Pure Functionals - LDA
    "lsda": "Local Spin Density Approximation (Slater exchange + VWN correlation)",
    
    # Pure Functionals - GGA
    "pbe": "Perdew-Burke-Ernzerhof (1996) GGA functional",
    "blyp": "Becke 1988 exchange + Lee-Yang-Parr correlation",
    "bp86": "Becke 1988 exchange + Perdew 1988 correlation",
    "b97-d3": "Grimme's 2006 B97 reparameterization with D3 dispersion",
    
    # Pure Functionals - Meta-GGA
    "r2scan": "Furness and Sun's 2020 r2SCAN meta-GGA functional",
    "tpss": "Tao-Perdew-Staroverov-Scuseria meta-GGA (2003)",
    "m06l": "Zhao and Truhlar's 2006 local meta-GGA functional",
    
    # Hybrid Functionals - Global
    "pbe0": "PBE0 hybrid functional (25% HF exchange)",
    "b3lyp": "B3LYP hybrid functional (20% HF exchange)",
    "b3pw91": "B3PW91 hybrid functional (20% HF exchange)",
    
    # Hybrid Functionals - Range-Separated
    "camb3lyp": "CAM-B3LYP range-separated hybrid (19-65% HF exchange)",
    "wb97x_d3": "ωB97X-D3 range-separated hybrid with D3 dispersion (20-100% HF exchange)",
    "wb97x_v": "ωB97X-V with VV10 nonlocal dispersion (17-100% HF exchange)",
    "wb97m_v": "ωB97M-V meta-GGA with VV10 dispersion (15-100% HF exchange)"
}

QC_BASIS_SETS = {
    # Pople's STO-nG minimal basis sets
    "sto-2g": "STO-2G minimal basis set",
    "sto-3g": "STO-3G minimal basis set (default if none specified)",
    "sto-4g": "STO-4G minimal basis set",
    "sto-5g": "STO-5G minimal basis set",
    "sto-6g": "STO-6G minimal basis set",
    
    # Pople's 6-31 basis sets (double-zeta)
    "6-31g": "6-31G split-valence double-zeta basis set",
    "6-31g*": "6-31G(d) - 6-31G with polarization on heavy atoms",
    "6-31g(d)": "6-31G with d polarization on heavy atoms",
    "6-31g(d,p)": "6-31G with polarization on all atoms",
    "6-31+g(d,p)": "6-31G with diffuse and polarization functions",
    "6-311+g(2d,p)": "6-311+G(2d,p) - larger basis for single-point calculations",
    
    # Jensen's pcseg-n basis sets (recommended for DFT)
    "pcseg-0": "Jensen pcseg-0 minimal basis set",
    "pcseg-1": "Jensen pcseg-1 double-zeta (better than 6-31G(d))",
    "pcseg-2": "Jensen pcseg-2 triple-zeta basis set",
    "pcseg-3": "Jensen pcseg-3 quadruple-zeta basis set",
    "pcseg-4": "Jensen pcseg-4 quintuple-zeta basis set",
    "aug-pcseg-1": "Augmented Jensen pcseg-1 double-zeta",
    "aug-pcseg-2": "Augmented Jensen pcseg-2 triple-zeta",
    
    # Dunning's cc-PVNZ basis sets (use seg-opt variants for speed)
    "cc-pvdz": "Correlation-consistent double-zeta (generally contracted - slow)",
    "cc-pvtz": "Correlation-consistent triple-zeta (generally contracted - slow)",
    "cc-pvqz": "Correlation-consistent quadruple-zeta (generally contracted - slow)",
    "cc-pvdz(seg-opt)": "cc-pVDZ segmented-optimized (preferred over cc-pVDZ)",
    "cc-pvtz(seg-opt)": "cc-pVTZ segmented-optimized (preferred over cc-pVTZ)",
    "cc-pvqz(seg-opt)": "cc-pVQZ segmented-optimized (preferred over cc-pVQZ)",
    
    # Ahlrichs's def2 basis sets
    "def2-sv(p)": "Ahlrichs def2-SV(P) split-valence polarized",
    "def2-svp": "Ahlrichs def2-SVP split-valence polarized",
    "def2-tzvp": "Ahlrichs def2-TZVP triple-zeta valence polarized",
    
    # Truhlar's efficient basis sets
    "midi!": "MIDI!/MIDIX polarized minimal basis set (very efficient)",
    "midix": "MIDI!/MIDIX polarized minimal basis set (very efficient)"
}

QC_CORRECTIONS = {
    "d3bj": "Grimme's D3 dispersion correction with Becke-Johnson damping",
    "d3": "Grimme's D3 dispersion correction (automatically applied for B97-D3, ωB97X-D3)"
}

def lookup_molecule_smiles(molecule_name: str) -> str:
    """Simple molecule name to SMILES lookup for common molecules."""
    common_molecules = {
        # Simple molecules
        "water": "O",
        "methane": "C",
        "ethane": "CC",
        "propane": "CCC",
        "butane": "CCCC",
        "ethylene": "C=C",
        "acetylene": "C#C",
        "benzene": "c1ccccc1",
        "toluene": "Cc1ccccc1",
        "phenol": "Oc1ccccc1",
        
        # Peroxides and simple radicals
        "hydrogen peroxide": "OO",
        "h2o2": "OO",
        "peroxide": "OO",
        
        # Common solvents
        "methanol": "CO",
        "ethanol": "CCO",
        "acetone": "CC(=O)C",
        "dmso": "CS(=O)C",
        "thf": "C1CCOC1",
    }
    
    # Check if it's already a SMILES string (contains specific characters)
    if any(char in molecule_name.lower() for char in ['=', '#', '[', '(', ')', '@']):
        return molecule_name
    
    # Look up common name
    lookup_name = molecule_name.lower().strip()
    smiles = common_molecules.get(lookup_name)
    
    if smiles:
        logger.info(f"Molecule lookup: '{molecule_name}' → '{smiles}'")
        return smiles
    else:
        return molecule_name

def rowan_scan(
    name: str,
    molecule: str,
    coordinate_type: str,
    atoms: Union[List[int], str],
    start: float,
    stop: float,
    num: int,
    method: Optional[str] = None,
    basis_set: Optional[str] = None,
    engine: Optional[str] = None,
    corrections: Optional[List[str]] = None,
    charge: int = 0,
    multiplicity: int = 1,
    mode: Optional[str] = None,
    constraints: Optional[List[Dict[str, Any]]] = None,
    # New parameters from ScanWorkflow
    coordinate_type_2d: Optional[str] = None,
    atoms_2d: Optional[Union[List[int], str]] = None,
    start_2d: Optional[float] = None,
    stop_2d: Optional[float] = None,
    num_2d: Optional[int] = None,
    wavefront_propagation: bool = True,
    concerted_coordinates: Optional[List[Dict[str, Any]]] = None,
    folder_uuid: Optional[str] = None,
    blocking: bool = True,
    ping_interval: int = 10
) -> str:
    """Run potential energy surface scans with full parameter control including 2D and concerted scans.
    
    Performs constrained optimizations along reaction coordinates using Rowan's
    wavefront propagation method to avoid local minima. Essential for:
    - Mapping reaction pathways and mechanisms
    - Finding transition state approximations  
    - Studying conformational preferences and rotational barriers
    - Analyzing atropisomerism and molecular flexibility
    - 2D potential energy surfaces
    - Concerted coordinate changes
    
    Args:
        name: Name for the scan calculation
        molecule: Molecule SMILES string or common name (e.g., "butane", "phenol")
        coordinate_type: Type of coordinate to scan - "bond", "angle", or "dihedral"
        atoms: List of 1-indexed atom numbers or comma-separated string (e.g., [1,2,3,4] or "1,2,3,4")
        start: Starting value of the coordinate (Å for bonds, degrees for angles/dihedrals)
        stop: Ending value of the coordinate
        num: Number of scan points to calculate
        method: QC method (default: "hf-3c" for speed, use "b3lyp" for accuracy)
        basis_set: Basis set (default: auto-selected based on method)
        engine: Computational engine (default: "psi4") 
        corrections: List of corrections like ["d3bj"] for dispersion
        charge: Molecular charge (default: 0)
        multiplicity: Spin multiplicity (default: 1 for singlet)
        mode: Calculation precision - "reckless", "rapid", "careful", "meticulous" (default: "rapid")
        constraints: Additional coordinate constraints during optimization
        coordinate_type_2d: Type of second coordinate for 2D scan (optional)
        atoms_2d: Atoms for second coordinate in 2D scan (optional)
        start_2d: Starting value of second coordinate (optional)
        stop_2d: Ending value of second coordinate (optional) 
        num_2d: Number of points for second coordinate (must equal num for 2D grid, optional)
        wavefront_propagation: Use wavefront propagation for smoother scans (default: True)
        concerted_coordinates: List of coordinate dictionaries for concerted scans (optional)
        folder_uuid: Optional folder UUID for organization
        blocking: Whether to wait for completion (default: True)
        ping_interval: Check status interval in seconds (default: 10, longer for scans)
    
    Returns:
        Scan results with energy profile and structural data
    """
    
    # Look up SMILES if a common name was provided
    canonical_smiles = lookup_molecule_smiles(molecule)
    
    # Validate coordinate type
    valid_coord_types = ["bond", "angle", "dihedral"]
    coord_type_lower = coordinate_type.lower()
    if coord_type_lower not in valid_coord_types:
        return f"Invalid coordinate_type '{coordinate_type}'. Valid types: {', '.join(valid_coord_types)}"
    
    # Handle string input for atoms (common format: "1,2,3,4")
    if isinstance(atoms, str):
        try:
            # Parse comma-separated string to list of integers
            atoms = [int(x.strip()) for x in atoms.split(",")]
        except ValueError as e:
            return f"Invalid atoms string '{atoms}'. Use format '1,2,3,4' or pass as list [1,2,3,4]. Error: {e}"
    
    # Ensure atoms is a list
    if not isinstance(atoms, list):
        return f"Atoms must be a list of integers or comma-separated string. Got: {type(atoms).__name__}"
    
    # Validate atom count for coordinate type
    expected_atoms = {"bond": 2, "angle": 3, "dihedral": 4}
    expected_count = expected_atoms[coord_type_lower]
    if len(atoms) != expected_count:
        return f"{coordinate_type} requires exactly {expected_count} atoms, got {len(atoms)}. Use format: [1,2,3,4] or '1,2,3,4'"
    
    # Validate atoms are positive integers
    if not all(isinstance(atom, int) and atom > 0 for atom in atoms):
        return f"All atom indices must be positive integers (1-indexed). Got: {atoms}. Use format: [1,2,3,4] or '1,2,3,4'"
    
    # Validate scan range
    if num < 2:
        return f"Number of scan points must be at least 2, got {num}"
    
    if start >= stop:
        return f"Start value ({start}) must be less than stop value ({stop})"
    
    # Handle 2D scan validation
    is_2d_scan = any([coordinate_type_2d, atoms_2d, start_2d is not None, stop_2d is not None, num_2d is not None])
    
    if is_2d_scan:
        # For 2D scan, all 2D parameters must be provided
        if not all([coordinate_type_2d, atoms_2d, start_2d is not None, stop_2d is not None, num_2d is not None]):
            return f"For 2D scans, all 2D parameters must be provided: coordinate_type_2d, atoms_2d, start_2d, stop_2d, num_2d"
        
        # Validate 2D coordinate type
        coord_type_2d_lower = coordinate_type_2d.lower()
        if coord_type_2d_lower not in valid_coord_types:
            return f"Invalid coordinate_type_2d '{coordinate_type_2d}'. Valid types: {', '.join(valid_coord_types)}"
        
        # Handle string input for 2D atoms
        if isinstance(atoms_2d, str):
            try:
                atoms_2d = [int(x.strip()) for x in atoms_2d.split(",")]
            except ValueError as e:
                return f"Invalid atoms_2d string '{atoms_2d}'. Use format '1,2,3,4' or pass as list [1,2,3,4]. Error: {e}"
        
        # Validate 2D atom count
        expected_count_2d = expected_atoms[coord_type_2d_lower]
        if len(atoms_2d) != expected_count_2d:
            return f"{coordinate_type_2d} requires exactly {expected_count_2d} atoms, got {len(atoms_2d)}"
        
        # Validate 2D atoms are positive integers
        if not all(isinstance(atom, int) and atom > 0 for atom in atoms_2d):
            return f"All 2D atom indices must be positive integers (1-indexed). Got: {atoms_2d}"
        
        # Validate 2D scan range
        if num_2d < 2:
            return f"Number of 2D scan points must be at least 2, got {num_2d}"
        
        if start_2d >= stop_2d:
            return f"2D start value ({start_2d}) must be less than stop value ({stop_2d})"
    
    # Handle concerted scan validation
    if concerted_coordinates:
        # Validate each coordinate in the concerted scan
        for i, coord in enumerate(concerted_coordinates):
            required_keys = {"coordinate_type", "atoms", "start", "stop", "num"}
            if not all(key in coord for key in required_keys):
                return f"Concerted coordinate {i+1} missing required keys: {required_keys}"
            
            # All concerted coordinates must have same number of steps
            if coord["num"] != num:
                return f"All concerted scan coordinates must have same number of steps. Got {coord['num']} vs {num}"
            
            # Validate coordinate type
            if coord["coordinate_type"].lower() not in valid_coord_types:
                return f"Invalid coordinate_type in concerted coordinate {i+1}: '{coord['coordinate_type']}'"
        
    # ENFORCE REQUIRED PARAMETERS FOR SCANWORKFLOW - Always provide robust defaults
    if method is None:
        method = "hf-3c"  # Fast, reliable default for scans
        
    if engine is None:
        engine = "psi4"  # Required by ScanWorkflow
        
    if mode is None:
        mode = "rapid"  # Good balance for scans
    
    # Ensure all required defaults are lowercase for API consistency
    method = method.lower()
    engine = engine.lower()
    mode = mode.lower()
        
    # Validate QC parameters if provided
    if method and method.lower() not in QC_METHODS and method.lower() != "hf-3c":
        available_methods = ", ".join(list(QC_METHODS.keys()) + ["hf-3c"])
        return f"Invalid method '{method}'. Available methods: {available_methods}"
    
    if basis_set and basis_set.lower() not in QC_BASIS_SETS:
        available_basis = ", ".join(QC_BASIS_SETS.keys())
        return f"Invalid basis set '{basis_set}'. Available basis sets: {available_basis}"
    
    if engine and engine.lower() not in QC_ENGINES:
        available_engines = ", ".join(QC_ENGINES.keys())
        return f"Invalid engine '{engine}'. Available engines: {available_engines}"
    
    if corrections:
        invalid_corrections = [corr for corr in corrections if corr.lower() not in QC_CORRECTIONS]
        if invalid_corrections:
            available_corrections = ", ".join(QC_CORRECTIONS.keys())
            return f"Invalid corrections {invalid_corrections}. Available corrections: {available_corrections}"
    
    # Validate mode
    valid_modes = ["reckless", "rapid", "careful", "meticulous"]
    if mode and mode.lower() not in valid_modes:
        return f"Invalid mode '{mode}'. Valid modes: {', '.join(valid_modes)}"
    
    # Log the scan parameters
    logger.info(f"Name: {name}")
    logger.info(f"Input: {molecule}")
    if is_2d_scan:
        logger.info(f"2D Grid Size: {num} × {num_2d} = {num * num_2d} total points")
    if concerted_coordinates:
        logger.info(f"Concerted scan with {len(concerted_coordinates) + 1} coordinates")
    logger.info(f"Mode: {mode}")
    logger.info(f"Wavefront Propagation: {wavefront_propagation}")
    
    try:
        # Build scan coordinate specification based on scan type
        scan_coordinates = []
        
        # BUILD PRIMARY SCAN_SETTINGS - EXACTLY MATCHING STJAMES SCANSETTINGS
        primary_coord = {
            "type": coord_type_lower,    # Required: matches stjames ScanSettings.type
            "atoms": atoms,              # Required: list of 1-indexed atom numbers
            "start": float(start),       # Required: starting coordinate value
            "stop": float(stop),         # Required: ending coordinate value  
            "num": int(num)              # Required: number of scan points
        }
        scan_coordinates.append(primary_coord)
        
        # ADD 2D COORDINATE IF SPECIFIED - EXACT STJAMES FORMAT
        if is_2d_scan:
            secondary_coord = {
                "type": coord_type_2d_lower,     # Required: matches stjames ScanSettings.type
                "atoms": atoms_2d,               # Required: list of 1-indexed atom numbers
                "start": float(start_2d),        # Required: starting coordinate value
                "stop": float(stop_2d),          # Required: ending coordinate value
                "num": int(num_2d)               # Required: number of scan points
            }
            scan_coordinates.append(secondary_coord)
        
        # ADD CONCERTED COORDINATES IF SPECIFIED - EXACT STJAMES FORMAT
        if concerted_coordinates:
            for i, coord in enumerate(concerted_coordinates):
                concerted_coord = {
                    "type": coord["coordinate_type"].lower(),   # Required: matches stjames ScanSettings.type
                    "atoms": coord["atoms"],                    # Required: list of 1-indexed atom numbers
                    "start": float(coord["start"]),            # Required: starting coordinate value
                    "stop": float(coord["stop"]),              # Required: ending coordinate value
                    "num": int(coord["num"])                   # Required: number of scan points
                }
                scan_coordinates.append(concerted_coord)
        
        # Build parameters for rowan.compute call
        compute_params = {
            "name": name,
            "molecule": canonical_smiles,  # Required by rowan.compute() API
            "folder_uuid": folder_uuid,
            "blocking": blocking,
            "ping_interval": ping_interval,
            # Add initial_molecule parameter for MoleculeWorkflow compatibility
            "initial_molecule": canonical_smiles  # Required by stjames MoleculeWorkflow
        }
        
        # SCAN_SETTINGS CONSTRUCTION - STRICTLY FOLLOWING STJAMES SCANWORKFLOW REQUIREMENTS
        if len(scan_coordinates) == 1:
            # SINGLE COORDINATE SCAN: scan_settings is a ScanSettings object
            compute_params["scan_settings"] = scan_coordinates[0]
        else:
            if is_2d_scan:
                # 2D SCAN: separate scan_settings and scan_settings_2d fields
                compute_params["scan_settings"] = scan_coordinates[0]      # Primary coordinate
                compute_params["scan_settings_2d"] = scan_coordinates[1]   # Secondary coordinate  
            else:
                # CONCERTED SCAN: scan_settings is a list of ScanSettings
                compute_params["scan_settings"] = scan_coordinates
        
        # Add wavefront propagation setting
        compute_params["wavefront_propagation"] = wavefront_propagation
        
        # BUILD REQUIRED CALC_SETTINGS - ALWAYS COMPLETE
        calc_settings = {
            "method": method,  # Always present due to defaults above
            "mode": mode,      # Always present due to defaults above
        }
        
        # Add charge/multiplicity only if non-default
        if charge != 0:
            calc_settings["charge"] = charge
        if multiplicity != 1:
            calc_settings["multiplicity"] = multiplicity
            
        # Add optional parameters if provided
        if basis_set:
            calc_settings["basis_set"] = basis_set.lower()
        if corrections:
            calc_settings["corrections"] = [corr.lower() for corr in corrections]
        if constraints:
            calc_settings["constraints"] = constraints
        
        # REQUIRED FIELDS - ScanWorkflow validation will fail without these
        compute_params["calc_settings"] = calc_settings
        compute_params["calc_engine"] = engine  # Always present due to defaults above
        
        # Submit the scan workflow
        result = rowan.compute(workflow_type="scan", **compute_params)
        
        # Format results based on status
        uuid = result.get('uuid', 'N/A')
        status = result.get('status', 'unknown')
        
        # Determine scan type for display
        scan_type = "1D Scan"
        total_points = num
        if is_2d_scan:
            scan_type = "2D Grid Scan"
            total_points = num * num_2d
        elif concerted_coordinates:
            scan_type = f"Concerted Scan ({len(concerted_coordinates) + 1} coordinates)"
            total_points = num
        
        # Format response
        if blocking:
            # Blocking mode - check if successful
            if status == "success":
                formatted = f"✅ {scan_type} '{name}' completed successfully!\n"
                formatted += f"🔖 Workflow UUID: {uuid}\n"
                formatted += f"📊 Status: {status}\n\n"
                
                # Extract scan results if available
                object_data = result.get("object_data", {})
                scan_points = object_data.get("scan_points", [])
                
                if scan_points:
                    formatted += f"📈 Scan Results: {len(scan_points)} points calculated\n"
                    
                    # Show first few scan points
                    for i, point in enumerate(scan_points[:3]):
                        if isinstance(point, dict):
                            energy = point.get("energy", "N/A")
                            formatted += f"   Point {i+1}: Energy = {energy}\n"
                    
                    if len(scan_points) > 3:
                        formatted += f"   ... and {len(scan_points) - 3} more points\n"
                else:
                    formatted += "📈 Results: Check workflow details for scan data\n"
                    
                return formatted
            else:
                # Failed calculation
                return f"❌ {scan_type} calculation failed\n🔖 UUID: {uuid}\n📋 Status: {status}\n💬 Check workflow details for more information"
        else:
            # Non-blocking mode - just submission confirmation
            formatted = f"📋 {scan_type} '{name}' submitted!\n"
            formatted += f"🔖 Workflow UUID: {uuid}\n"
            formatted += f"⏳ Status: Running...\n"
            formatted += f"💡 Use rowan_workflow_management to check status\n\n"
            
            formatted += f"Scan Details:\n"
            formatted += f"🧬 Molecule: {canonical_smiles}\n"
            formatted += f"📐 Coordinate: {coordinate_type.upper()} on atoms {atoms} ({start} to {stop}, {num} points)\n"
            if is_2d_scan:
                formatted += f"📐 Secondary: {coordinate_type_2d.upper()} on atoms {atoms_2d} ({start_2d} to {stop_2d}, {num_2d} points)\n"
            formatted += f"📊 Total Points: {total_points}\n"
            formatted += f"⚙️  Method: {method}, Engine: {engine}\n"
            
            return formatted
            
    except Exception as e:
        logger.error(f"Error in rowan_scan: {str(e)}")
        return f"PES scan submission failed: {str(e)}"

def test_rowan_scan():
    """Test the rowan_scan function."""
    try:
        # Test with minimal parameters - should use all defaults
        result = rowan_scan(
            name="test_scan_ethane",
            molecule="CC",
            coordinate_type="bond",
            atoms=[1, 2],
            start=1.4,
            stop=1.8,
            num=3,  # Very short for testing
            blocking=False
        )
        print("✅ Scan test successful!")
        print(f"Result: {result[:200]}..." if len(result) > 200 else result)
        return True
    except Exception as e:
        print(f"❌ Scan test failed: {e}")
        return False

if __name__ == "__main__":
    test_rowan_scan()


"""
Molecule name to SMILES converter using Chemical Identifier Resolver (CIR).
Enables natural language molecule input for Rowan workflows.
"""

from typing import List, Dict, Annotated
from urllib.request import urlopen
from urllib.parse import quote
import logging

logger = logging.getLogger(__name__)


def molecule_lookup(
    molecule_name: Annotated[str, "Common name, IUPAC name, or CAS number of molecule (e.g., 'aspirin', 'caffeine', '50-78-2')"],
    fallback_to_input: Annotated[bool, "If lookup fails, return the input string assuming it might be SMILES"] = False
) -> str:
    """Convert molecule names to SMILES using Chemical Identifier Resolver (CIR).
    
    Args:
        molecule_name: Common name, IUPAC name, or CAS number of molecule (e.g., 'aspirin', 'caffeine', '50-78-2')
        fallback_to_input: If lookup fails, return the input string assuming it might be SMILES
    
    This tool enables natural language input for molecules by converting common names,
    IUPAC names, CAS numbers, and other identifiers to SMILES strings that can be
    used with Rowan workflows.
    
    Supported Input Types:
    - Common names: 'aspirin', 'caffeine', 'benzene', 'glucose'
    - IUPAC names: '2-acetoxybenzoic acid', '1,3,7-trimethylpurine-2,6-dione'
    - CAS numbers: '50-78-2' (aspirin), '58-08-2' (caffeine)
    - InChI strings
    - Already valid SMILES (will be validated)
    
    Returns:
        SMILES string if successful, error message if not found
        
    Examples:
        # Common drug name
        result = molecule_lookup("aspirin")
        # Returns: "CC(=O)Oc1ccccc1C(=O)O"
        
        # IUPAC name
        result = molecule_lookup("2-acetoxybenzoic acid")
        # Returns: "CC(=O)Oc1ccccc1C(=O)O"
        
        # CAS number
        result = molecule_lookup("50-78-2")
        # Returns: "CC(=O)Oc1ccccc1C(=O)O"
        
        # Complex molecule
        result = molecule_lookup("paracetamol")
        # Returns: "CC(=O)Nc1ccc(O)cc1"
    """
    try:
        # Clean input
        molecule_name = molecule_name.strip()
        
        # Check if already SMILES-like (contains typical SMILES characters)
        smiles_chars = {'=', '#', '(', ')', '[', ']', '@', '+', '-'}
        if any(char in molecule_name for char in smiles_chars):
            logger.info(f"Input '{molecule_name}' appears to be SMILES, returning as-is")
            return molecule_name
        
        # Query CIR service
        logger.info(f"Looking up molecule: {molecule_name}")
        url = f'http://cactus.nci.nih.gov/chemical/structure/{quote(molecule_name)}/smiles'
        
        response = urlopen(url, timeout=10)
        smiles = response.read().decode('utf8').strip()
        
        # CIR may return multiple SMILES for some queries, take the first one
        if '\n' in smiles:
            smiles = smiles.split('\n')[0]
        
        logger.info(f"Successfully converted '{molecule_name}' to SMILES: {smiles}")
        return smiles
        
    except Exception as e:
        logger.warning(f"Failed to lookup '{molecule_name}': {e}")
        
        if fallback_to_input:
            logger.info(f"Returning original input as fallback: {molecule_name}")
            return molecule_name
        else:
            return f"Could not find SMILES for '{molecule_name}'. Please check the name or provide a valid SMILES string."


def batch_molecule_lookup(
    molecule_names: Annotated[List[str], "List of molecule names to convert to SMILES"],
    skip_failures: Annotated[bool, "Skip molecules that fail lookup instead of stopping"] = True
) -> Dict[str, str]:
    """Convert multiple molecule names to SMILES in batch.
    
    Args:
        molecule_names: List of molecule names to convert to SMILES
        skip_failures: Skip molecules that fail lookup instead of stopping
    
    Useful for preparing multiple molecules for workflows or screening.
        
    Returns:
        Dictionary mapping input names to SMILES strings (or error messages)
        
    Examples:
        # Drug screening set
        result = batch_molecule_lookup([
            "aspirin",
            "ibuprofen", 
            "paracetamol",
            "caffeine"
        ])
        # Returns: {
        #     "aspirin": "CC(=O)Oc1ccccc1C(=O)O",
        #     "ibuprofen": "CC(C)Cc1ccc(C(C)C(=O)O)cc1",
        #     "paracetamol": "CC(=O)Nc1ccc(O)cc1",
        #     "caffeine": "CN1C=NC2=C1C(=O)N(C(=O)N2C)C"
        # }
        
        # Mixed input types
        result = batch_molecule_lookup([
            "benzene",           # Common name
            "50-78-2",          # CAS number
            "ethanoic acid"     # IUPAC name
        ])
    """
    results = {}
    
    for name in molecule_names:
        try:
            smiles = molecule_lookup(name, fallback_to_input=False)
            results[name] = smiles
        except Exception as e:
            error_msg = f"Lookup failed: {str(e)}"
            if skip_failures:
                logger.warning(f"Skipping {name}: {error_msg}")
                results[name] = error_msg
            else:
                raise ValueError(f"Failed to lookup '{name}': {error_msg}")
    
    return results


def validate_smiles(
    smiles: Annotated[str, "SMILES string to validate"]
) -> Dict[str, any]:
    """Validate a SMILES string and return basic molecular properties.
    
    Args:
        smiles: SMILES string to validate
    
    Uses RDKit to validate SMILES and extract basic properties.
        
    Returns:
        Dictionary with validation status and properties if valid
        
    Examples:
        result = validate_smiles("CC(=O)O")
        # Returns: {
        #     "valid": True,
        #     "canonical_smiles": "CC(=O)O",
        #     "molecular_formula": "C2H4O2",
        #     "molecular_weight": 60.05
        # }
    """
    try:
        from rdkit import Chem
        from rdkit.Chem import Descriptors
        
        mol = Chem.MolFromSmiles(smiles)
        
        if mol is None:
            return {
                "valid": False,
                "error": "Invalid SMILES string"
            }
        
        return {
            "valid": True,
            "canonical_smiles": Chem.MolToSmiles(mol),
            "molecular_formula": Chem.rdMolDescriptors.CalcMolFormula(mol),
            "molecular_weight": round(Descriptors.MolWt(mol), 2),
            "num_atoms": mol.GetNumAtoms(),
            "num_bonds": mol.GetNumBonds()
        }
        
    except ImportError:
        return {
            "valid": "unknown",
            "error": "RDKit not available for validation"
        }
    except Exception as e:
        return {
            "valid": False,
            "error": str(e)
        }


# Common molecules reference (for documentation)
COMMON_MOLECULES = {
    # Drugs
    "aspirin": "CC(=O)Oc1ccccc1C(=O)O",
    "paracetamol": "CC(=O)Nc1ccc(O)cc1",
    "acetaminophen": "CC(=O)Nc1ccc(O)cc1",  # Same as paracetamol
    "ibuprofen": "CC(C)Cc1ccc(C(C)C(=O)O)cc1",
    "caffeine": "CN1C=NC2=C1C(=O)N(C(=O)N2C)C",
    "penicillin": "CC1(C)SC2C(NC(=O)Cc3ccccc3)C(=O)N2C1C(=O)O",
    
    # Solvents
    "water": "O",
    "ethanol": "CCO",
    "methanol": "CO",
    "acetone": "CC(=O)C",
    "dmso": "CS(=O)C",
    "chloroform": "C(Cl)(Cl)Cl",
    "benzene": "c1ccccc1",
    "toluene": "Cc1ccccc1",
    
    # Organic compounds
    "glucose": "C(C1C(C(C(C(O1)O)O)O)O)O",
    "acetic acid": "CC(=O)O",
    "ethanoic acid": "CC(=O)O",  # IUPAC for acetic acid
    "phenol": "Oc1ccccc1",
    "aniline": "Nc1ccccc1",
    "naphthalene": "c1ccc2c(c1)cccc2",
    
    # Amino acids
    "glycine": "C(C(=O)O)N",
    "alanine": "CC(C(=O)O)N",
    "valine": "CC(C)C(C(=O)O)N",
    "leucine": "CC(C)CC(C(=O)O)N",
    "lysine": "C(CCN)CC(C(=O)O)N",
}


"""
Calculate pKa values for molecules using Rowan API.
"""

import os
import rowan
from typing import Optional

# Set up logging
import logging
logger = logging.getLogger(__name__)



def log_rowan_api_call(workflow_type: str, **kwargs):
    """Log Rowan API calls and let Rowan handle its own errors."""
    
    # Simple logging for calculations
    logger.info(f" Starting {workflow_type.replace('_', ' ')}...")
    
    # Let Rowan handle everything - no custom error handling
    return rowan.compute(workflow_type=workflow_type, **kwargs)

def rowan_pka(
    name: str,
    molecule: str,
    folder_uuid: Optional[str] = None,
    blocking: bool = True,
    ping_interval: int = 5
) -> str:
    """Calculate pKa values for molecules.
    
    Args:
        name: Name for the calculation
        molecule: Molecule SMILES string
        folder_uuid: UUID of folder to organize calculation in
        blocking: Whether to wait for completion (default: True)
        ping_interval: How often to check status in seconds (default: 5)
    
    Returns:
        pKa calculation results
    """
    try:
        result = log_rowan_api_call(
            workflow_type="pka",
            name=name,
            molecule=molecule,
            folder_uuid=folder_uuid,
            blocking=blocking,
            ping_interval=ping_interval
        )
        
        return str(result)
        
    except Exception as e:
        error_response = {
            "error": f"pKa calculation failed: {str(e)}",
            "name": name,
            "molecule": molecule
        }
        return str(error_response)


def test_rowan_pka():
    """Test the rowan_pka function."""
    try:
        # Test with minimal parameters
        result = rowan_pka(
            name="test_pka_water",
            molecule="O"
        )
        print("pKa test successful")
        print(f"Result: {result}")
        return True
    except Exception as e:
        print(f"pKa test failed: {e}")
        return False


if __name__ == "__main__":
    test_rowan_pka() 


"""
Rowan docking function for MCP tool integration.
Implements protein-ligand docking following the stjames-public workflow pattern.
"""

from typing import Optional, Union, List, Tuple, Dict, Any
import rowan
import logging
import os
import requests

# Set up logging
logger = logging.getLogger(__name__)

# Get API key from environment
api_key = os.environ.get("ROWAN_API_KEY")
if api_key:
    rowan.api_key = api_key
else:
    logger.warning("ROWAN_API_KEY not found in environment")

def fetch_pdb_content(pdb_id: str) -> Optional[str]:
    """Fetch PDB content from RCSB PDB database.
    
    Args:
        pdb_id: 4-character PDB ID (e.g., "1COX")
        
    Returns:
        PDB file content as string, or None if fetch fails
    """
    try:
        url = f"https://files.rcsb.org/download/{pdb_id.upper()}.pdb"
        response = requests.get(url, timeout=30)
        response.raise_for_status()
        return response.text
    except Exception as e:
        logger.error(f"Failed to fetch PDB {pdb_id}: {str(e)}")
        return None

def rowan_docking(
    name: str,
    # Ligand specification (following stjames priority)
    molecules: Optional[List[str]] = None,
    smiles: Optional[Union[str, List[str]]] = None,
    # Protein specification  
    target: Optional[str] = None,
    target_uuid: Optional[str] = None,
    # Pocket specification
    pocket: Optional[Tuple[Tuple[float, float, float], Tuple[float, float, float]]] = None,
    # Workflow parameters
    do_csearch: bool = True,
    do_optimization: bool = True,
    do_pose_refinement: bool = True,
    conformers: Optional[List[str]] = None,
    # MCP workflow parameters
    folder_uuid: Optional[str] = None,
    blocking: bool = True,
    ping_interval: int = 5
) -> str:
    """Perform protein-ligand docking using Rowan's ML workflow.
    
    Performs comprehensive protein-ligand docking with conformer generation,
    strain filtering, AutoDock Vina docking, and pose refinement.
    
    IMPORTANT: For docking, you need to provide the protein target in one of two ways:
    1. target_uuid: UUID of a pre-uploaded protein in your Rowan account
    2. target: Reserved for future use (currently not supported via API)
    
    To dock against a known protein like 1COX, you should:
    1. Upload the PDB file to your Rowan account via the web interface
    2. Get the protein UUID from your account
    3. Use that UUID with the target_uuid parameter
    
    Args:
        name: Name for the docking calculation
        molecules: List of molecules to dock (SMILES strings)
        smiles: SMILES string(s) of ligands (can be string or list)
        target: Not currently supported - use target_uuid instead
        target_uuid: UUID of pre-uploaded protein target from your Rowan account
        pocket: Tuple of (center, size) where each is (x, y, z) coordinates
                Example: ((10.0, 15.0, 20.0), (20.0, 20.0, 20.0))
        do_csearch: Whether to perform conformer search (default: True)
        do_optimization: Whether to optimize starting structures (default: True)
        do_pose_refinement: Whether to optimize non-rotatable bonds in output poses (default: True)
        conformers: List of pre-optimized conformer UUIDs (optional)
        folder_uuid: Optional folder UUID to organize the calculation
        blocking: Whether to wait for completion (default: True)
        ping_interval: Interval in seconds to check calculation status
        
    Returns:
        Formatted string with workflow results or submission confirmation
        
    Examples:
        # Using pre-uploaded protein
        rowan_docking(
            name="aspirin_docking",
            smiles="CC(=O)Oc1ccccc1C(=O)O",
            target_uuid="your-protein-uuid-here",
            pocket=((10.0, 15.0, 20.0), (20.0, 20.0, 20.0))
        )
        
        # Multiple molecules
        rowan_docking(
            name="multi_ligand_docking",
            molecules=["CCO", "CC(C)O", "CCCO"],
            target_uuid="your-protein-uuid-here",
            pocket=((0.0, 0.0, 0.0), (25.0, 25.0, 25.0))
        )
    """
    
    # Validate that we have ligand input
    if molecules is None and smiles is None:
        return "Error: Either 'molecules' or 'smiles' parameter must be provided"
    
    # Validate that we have protein input
    if target is None and target_uuid is None:
        return "Error: Either 'target' or 'target_uuid' parameter must be provided"
    
    # If target is provided, inform user it's not currently supported
    if target is not None:
        return ("Error: Direct PDB content upload is not currently supported through this API. "
                "Please use one of these alternatives:\n"
                "1. Upload your PDB file through the Rowan web interface at https://labs.rowansci.com\n"
                "2. Get the protein UUID from your account\n"
                "3. Use the 'target_uuid' parameter with that UUID\n\n"
                "For known proteins like 1COX, 2COX, etc., they may already be available "
                "in the Rowan protein library.")
    
    # Validate pocket if provided
    if pocket is not None:
        try:
            center, size = pocket
            if len(center) != 3 or len(size) != 3:
                return "Error: Pocket must be ((x,y,z), (x,y,z)) format"
            
            # Validate that size dimensions are positive
            if any(s <= 0 for s in size):
                return "Error: Pocket size dimensions must be positive"
                
        except (TypeError, ValueError) as e:
            return f"Error: Invalid pocket format. Expected ((x,y,z), (x,y,z)). Got: {pocket}"
    else:
        # Default pocket if not provided
        pocket = ((0.0, 0.0, 0.0), (20.0, 20.0, 20.0))
        logger.info("Using default pocket: center=(0,0,0), size=(20,20,20)")
    
    # Determine initial_molecule for rowan.compute()
    # For docking, the 'molecule' parameter is the ligand to dock
    initial_molecule = None
    ligands_list = None
    
    if molecules and len(molecules) > 0:
        initial_molecule = molecules[0]
        ligands_list = molecules
    elif isinstance(smiles, str):
        initial_molecule = smiles
        ligands_list = [smiles]
    elif isinstance(smiles, list) and len(smiles) > 0:
        initial_molecule = smiles[0]
        ligands_list = smiles
    else:
        return "Error: Could not determine initial molecule from inputs"
    
    try:
        # Build parameters following stjames DockingWorkflow structure
        compute_params = {
            "name": name,
            "molecule": initial_molecule,  # Required by rowan.compute() - this is the ligand
            "workflow_type": "docking",
            "folder_uuid": folder_uuid,
            "blocking": blocking,
            "ping_interval": ping_interval,
            # DockingWorkflow specific parameters
            "do_csearch": do_csearch,
            "do_optimization": do_optimization,
            "do_pose_refinement": do_pose_refinement,
            "pocket": pocket,
        }
        
        # For multiple ligands, pass them as 'molecules' parameter
        # This allows docking multiple ligands in one workflow
        if ligands_list and len(ligands_list) > 1:
            compute_params["molecules"] = ligands_list
            
        # Add protein specification (only one of these should be present)
        if target_uuid is not None:
            compute_params["target_uuid"] = target_uuid
        # Note: Direct PDB upload is not supported in current implementation
        # Users must pre-upload proteins through the Rowan web interface
            
        # Add conformers if provided
        if conformers is not None:
            compute_params["conformers"] = conformers
        
        # Submit docking calculation and return raw result
        result = rowan.compute(**compute_params)
        return result
            
    except Exception as e:
        logger.error(f"Error in rowan_docking: {str(e)}")
        return f"Docking calculation failed: {str(e)}"

def rowan_docking_pdb_id(
    name: str,
    pdb_id: str,
    # Ligand specification
    molecules: Optional[List[str]] = None,
    smiles: Optional[Union[str, List[str]]] = None,
    # Other parameters same as rowan_docking
    pocket: Optional[Tuple[Tuple[float, float, float], Tuple[float, float, float]]] = None,
    do_csearch: bool = True,
    do_optimization: bool = True,
    do_pose_refinement: bool = True,
    conformers: Optional[List[str]] = None,
    folder_uuid: Optional[str] = None,
    blocking: bool = True,
    ping_interval: int = 5
) -> str:
    """Helper function that explains how to dock with a PDB ID.
    
    NOTE: Direct PDB upload is not currently supported through the API.
    This function provides guidance on how to dock against known proteins.
    
    Args:
        name: Name for the docking calculation
        pdb_id: 4-character PDB ID (e.g., "1COX", "2COX", "6COX")
        (other parameters same as rowan_docking)
        
    Returns:
        Instructions on how to perform docking with the given PDB
    """
    # Check if we can fetch the PDB to verify it exists
    logger.info(f"Checking if PDB {pdb_id} exists...")
    pdb_content = fetch_pdb_content(pdb_id)
    
    if pdb_content is None:
        return (f"Error: PDB ID '{pdb_id}' not found in RCSB database. "
                f"Please check that the PDB ID is valid.")
    
    # Provide instructions for the user
    ligand_param = ""
    if molecules:
        ligand_param = f"    molecules={molecules},\n"
    elif smiles:
        if isinstance(smiles, str):
            ligand_param = f'    smiles="{smiles}",\n'
        else:
            ligand_param = f"    smiles={smiles},\n"
    
    return (f"PDB {pdb_id} found in RCSB database!\n\n"
            f"To perform docking with this protein:\n\n"
            f"1. Go to https://labs.rowansci.com\n"
            f"2. Upload the PDB file for {pdb_id}\n"
            f"   - You can download it from: https://files.rcsb.org/download/{pdb_id.upper()}.pdb\n"
            f"3. Once uploaded, find the protein UUID in your account\n"
            f"4. Use the docking function with target_uuid:\n\n"
            f"```python\n"
            f"rowan_docking(\n"
            f"    name=\"{name}\",\n"
            f"{ligand_param}"
            f"    target_uuid=\"your-protein-uuid-here\",\n"
            f"    pocket={pocket},\n"
            f"    do_csearch={do_csearch},\n"
            f"    do_optimization={do_optimization},\n"
            f"    do_pose_refinement={do_pose_refinement}\n"
            f")\n"
            f"```\n\n"
            f"Note: Some common proteins may already be available in the Rowan protein library.\n"
            f"Check your account for pre-uploaded structures.")

def test_rowan_docking():
    """Test the rowan_docking function."""
    try:
        # Test with minimal parameters
        result = rowan_docking(
            name="test_docking",
            smiles="CCO",  # Ethanol
            target_uuid="test-protein-uuid",  # Would need real UUID
            pocket=((0.0, 0.0, 0.0), (20.0, 20.0, 20.0)),
            blocking=False
        )
        print("Docking test result:")
        print(result)
        return True
    except Exception as e:
        print(f"Docking test failed: {e}")
        return False

if __name__ == "__main__":
    test_rowan_docking()


"""
Run an optimization calculation on a molecule using Rowan.

See documentation at: https://docs.rowansci.com/science/quantum-chemistry/geometry-optimization
"""

from stjames import Molecule

import rowan

# rowan.api_key = ""

result = rowan.submit_basic_calculation_workflow(
    initial_molecule=Molecule.from_smiles("O"),
    method="GFN2-xTB",
    tasks=["optimize"],
    engine="xtb",
    name="Water Optimization",
)

print(result)


import rowan

# rowan.api_key = ""

# Run calculation remotely
result = rowan.submit_protein_cofolding_workflow(
    initial_protein_sequences=[
        "ASKGTSHEAGIVCRITKPALLVLNHETAKVIQTAFQRASYPDITGEKAMMLLGQVKYGLHNIQISHLSIASSQVELVEAKSIDVSIQDVSVVFKGTLKYGYTTAWWLGIDQSIDFEIDSAIDLQINTQLTADSGRVRTDAPDCYLSFHKLLLHLQGEREPGWIKQLFTNFISFTLKLVLKGQICKEINVISNIMADFVQTRAASILSDGDIGVDISLTGDPVITASYLESHHKGHFIYKDVSEDLPLPTFSPTLLGDSRMLYFWFSERVFHSLAKVAFQDGRLMLSLMGDEFKAVLETWGFNTNQEIFQEVVGGFPSQAQVTVHCLKMPKISCQNKGVVVDSSVMVKFLFPRPDQQHSVAYTFEEDIVTTVQASYSKKKLFLSLLDFQITPKTVSNLTESSSESIQSFLQSMITAVGIPEVMSRLEVVFTALMNSKGVSLFDIINPEIITRDGFLLLQMDFGFPEHLLVDFLQSLS"
    ],
    initial_smiles_list=[
        "CCOC(=O)N1c2ccc(C(F)(F)F)cc2[C@@H](N(Cc2cc(C(F)(F)F)cc(C(F)(F)F)c2)C(=O)OC)C[C@H]1CC"
    ],
    ligand_binding_affinity_index=0,
    name="Torcetrapib Cofolding",
)

print(result)


"""MacropKa workflow function for MCP server."""

import os
import logging
from typing import Optional, Union, List

import rowan

# Configure logging
logger = logging.getLogger(__name__)

# Get API key from environment
api_key = os.environ.get("ROWAN_API_KEY")
if api_key:
    rowan.api_key = api_key
else:
    logger.warning("ROWAN_API_KEY not found in environment")


def log_rowan_api_call(func_name: str, **kwargs):
    """Log Rowan API calls for debugging."""
    logger.debug(f"Calling {func_name} with args: {kwargs}")


def rowan_macropka(
    name: str,
    molecule: str,
    min_pH: float = 0.0,
    max_pH: float = 14.0,
    max_charge: int = 2,
    min_charge: int = -2,
    compute_aqueous_solubility: bool = False,
    compute_solvation_energy: bool = True,
    folder_uuid: Optional[str] = None,
    blocking: bool = True,
    ping_interval: int = 5
) -> str:
    """
    Calculate macroscopic pKa values and related properties for a molecule.
    
    This workflow computes pKa values, microstates, isoelectric point, and optionally
    solvation energy and aqueous solubility across different pH values.
    
    Args:
        name: Name for the calculation
        molecule: SMILES string of the molecule
        min_pH: Minimum pH for calculations (default: 0.0)
        max_pH: Maximum pH for calculations (default: 14.0)
        max_charge: Maximum charge to consider for microstates (default: 2)
        min_charge: Minimum charge to consider for microstates (default: -2)
        compute_aqueous_solubility: Whether to compute aqueous solubility by pH (default: False)
        compute_solvation_energy: Whether to compute solvation energy for Kpuu (default: True)
        folder_uuid: UUID of folder to save results in
        blocking: Wait for calculation to complete (default: True)
        ping_interval: How often to check status in blocking mode (default: 5 seconds)
        
    Returns:
        String with workflow UUID or results depending on blocking mode
    """
    try:
        # Validate pH range
        if min_pH >= max_pH:
            return "Error: min_pH must be less than max_pH"
        
        # Validate charge range  
        if min_charge >= max_charge:
            return "Error: min_charge must be less than max_charge"
            
        # Log the API call
        log_rowan_api_call(
            "rowan.compute",
            workflow_type="macropka",
            name=name,
            molecule=molecule,
            min_pH=min_pH,
            max_pH=max_pH,
            max_charge=max_charge,
            min_charge=min_charge,
            compute_aqueous_solubility=compute_aqueous_solubility,
            compute_solvation_energy=compute_solvation_energy,
            folder_uuid=folder_uuid,
            blocking=blocking,
            ping_interval=ping_interval
        )
        
        # Submit calculation
        result = rowan.compute(
            workflow_type="macropka",
            name=name,
            molecule=molecule,  # Required by rowan.compute() API
            folder_uuid=folder_uuid,
            blocking=blocking,
            ping_interval=ping_interval,
            # Workflow-specific parameters for MacropKaWorkflow
            initial_smiles=molecule,  # Required by MacropKaWorkflow Pydantic model
            min_pH=min_pH,
            max_pH=max_pH,
            max_charge=max_charge,
            min_charge=min_charge,
            compute_aqueous_solubility=compute_aqueous_solubility,
            compute_solvation_energy=compute_solvation_energy
        )
        
        return result
            
    except Exception as e:
        logger.error(f"Error in rowan_macropka: {str(e)}")
        return f"MacropKa calculation failed: {str(e)}"


# Test function
if __name__ == "__main__":
    # Test with ethanol
    result = rowan_macropka(
        name="Ethanol MacropKa Test",
        molecule="CCO",
        compute_aqueous_solubility=True,
        blocking=True
    )
    print(result)


"""
Rowan v2 API: MacropKa Workflow
Calculate macroscopic pKa values across a pH range.
"""

from typing import Annotated
import rowan

def submit_macropka_workflow(
    initial_smiles: Annotated[str, "SMILES string of the molecule for macropKa calculation"],
    min_pH: Annotated[int, "Minimum pH value for the calculation range"] = 0,
    max_pH: Annotated[int, "Maximum pH value for the calculation range"] = 14,
    min_charge: Annotated[int, "Minimum molecular charge to consider"] = -2,
    max_charge: Annotated[int, "Maximum molecular charge to consider"] = 2,
    compute_solvation_energy: Annotated[bool, "Whether to compute solvation energy corrections"] = True,
    name: Annotated[str, "Workflow name for identification and tracking"] = "Macropka Workflow",
    folder_uuid: Annotated[str, "UUID of folder to organize this workflow. Empty string uses default folder"] = "",
    max_credits: Annotated[int, "Maximum credits to spend on this calculation. 0 for no limit"] = 0
):
    """Submit a MacropKa workflow using Rowan v2 API.
    
    Args:
        initial_smiles: SMILES string of the molecule for macropKa calculation
        min_pH: Minimum pH value for the calculation range
        max_pH: Maximum pH value for the calculation range
        min_charge: Minimum molecular charge to consider
        max_charge: Maximum molecular charge to consider
        compute_solvation_energy: Whether to compute solvation energy for each species
        name: Workflow name for identification and tracking
        folder_uuid: UUID of folder to organize this workflow. Empty string uses default folder.
        max_credits: Maximum credits to spend on this calculation. 0 for no limit.
    
    Calculates macroscopic pKa values across a pH range, determining the
    protonation states and their populations at different pH values.
    
    The workflow will:
    1. Identify all ionizable sites in the molecule
    2. Calculate microscopic pKa values for each site
    3. Determine macroscopic pKa values and species populations
    4. Optionally compute solvation energies
    
    Returns:
        Workflow object representing the submitted workflow
        
    Examples:
        # Simple molecule macropKa
        result = submit_macropka_workflow(
            initial_smiles="CC(=O)O",  # Acetic acid
            min_pH=0,
            max_pH=14
        )
        
        # Complex molecule with custom charge range
        result = submit_macropka_workflow(
            initial_smiles="CC(C)CC(C(=O)O)N",  # Leucine
            min_pH=0,
            max_pH=14,
            min_charge=-3,
            max_charge=3,
            compute_solvation_energy=True
        )
    """
    
    try:
        # Build workflow_data
        workflow_data = {
            "min_pH": min_pH,
            "max_pH": max_pH,
            "min_charge": min_charge,
            "max_charge": max_charge,
            "compute_solvation_energy": compute_solvation_energy,
        }
        
        # Build the API request
        data = {
            "name": name,
            "folder_uuid": folder_uuid,
            "workflow_type": "macropka",
            "workflow_data": workflow_data,
            "initial_smiles": initial_smiles,
            "max_credits": max_credits,
        }
        
        # Submit to API using rowan module
        return rowan.submit_macropka_workflow(
            initial_smiles=initial_smiles,
            min_pH=min_pH,
            max_pH=max_pH,
            min_charge=min_charge,
            max_charge=max_charge,
            compute_solvation_energy=compute_solvation_energy,
            name=name,
            folder_uuid=folder_uuid if folder_uuid else None,
            max_credits=max_credits if max_credits > 0 else None
        )
            
    except Exception as e:
        # Re-raise the exception so MCP can handle it
        raise e


"""
PDB handler for Rowan docking - attempts to handle PDB input intelligently
"""

import os
import logging
import requests
from typing import Optional, Dict, Any, Tuple
import rowan

logger = logging.getLogger(__name__)

def parse_pdb_content(pdb_content: str) -> Dict[str, Any]:
    """
    Parse PDB content into a structured format that might work with Rowan.
    
    This attempts to extract key information from PDB files that the 
    Rowan API might accept.
    """
    lines = pdb_content.split('\n')
    
    # Extract basic information
    header = ""
    title = ""
    atoms = []
    
    for line in lines:
        if line.startswith("HEADER"):
            header = line[10:50].strip()
        elif line.startswith("TITLE"):
            title += line[10:].strip() + " "
        elif line.startswith("ATOM") or line.startswith("HETATM"):
            atoms.append(line)
    
    # Try to create a structure that might work with Rowan's PDB format
    pdb_data = {
        "description": {
            "title": title.strip() or header,
            "classification": header
        },
        "experiment": {
            "technique": "X-RAY DIFFRACTION"  # Default assumption
        },
        "geometry": {},
        "contents": pdb_content,  # Keep full content
        "name": f"pdb_upload_{len(pdb_content)}"
    }
    
    return pdb_data

def attempt_pdb_upload(pdb_id: str, pdb_content: str, folder_uuid: Optional[str] = None) -> Tuple[Optional[str], str]:
    """
    Attempt to upload PDB content to Rowan.
    
    Returns:
        Tuple of (uuid or None, status message)
    """
    try:
        # First, try the Protein.create method if it exists
        if hasattr(rowan, 'Protein') and hasattr(rowan.Protein, 'create'):
            logger.info(f"Attempting to create protein from PDB {pdb_id}")
            result = rowan.Protein.create(
                name=f"{pdb_id}_upload",
                pdb_content=pdb_content,
                folder_uuid=folder_uuid
            )
            if result and 'uuid' in result:
                return result['uuid'], f"Successfully uploaded PDB {pdb_id}"
        
        # Try alternative upload methods
        if hasattr(rowan, 'upload_pdb'):
            result = rowan.upload_pdb(pdb_content, name=pdb_id)
            if result and 'uuid' in result:
                return result['uuid'], f"Successfully uploaded PDB {pdb_id}"
        
        # Try to create a file upload
        if hasattr(rowan, 'File') and hasattr(rowan.File, 'create'):
            result = rowan.File.create(
                name=f"{pdb_id}.pdb",
                content=pdb_content,
                file_type="pdb",
                folder_uuid=folder_uuid
            )
            if result and 'uuid' in result:
                return result['uuid'], f"Uploaded PDB {pdb_id} as file"
                
    except Exception as e:
        logger.error(f"Failed to upload PDB {pdb_id}: {str(e)}")
    
    return None, f"Cannot upload PDB through API - manual upload required"

def handle_pdb_input(pdb_input: str, folder_uuid: Optional[str] = None) -> Dict[str, Any]:
    """
    Handle PDB input (either PDB ID or UUID) and return information for docking.
    
    Args:
        pdb_input: Either a PDB ID (like "4Z18") or a UUID
        folder_uuid: Optional folder for uploads
        
    Returns:
        Dictionary with:
        - success: bool
        - uuid: str (if successful)
        - message: str (status/error message)
        - pdb_id: str (if it was a PDB ID)
        - instructions: str (what to do next)
    """
    # Check if it's a UUID (36 chars with dashes)
    if len(pdb_input) == 36 and pdb_input.count('-') == 4:
        return {
            "success": True,
            "uuid": pdb_input,
            "message": "Using provided protein UUID",
            "instructions": None
        }
    
    # Assume it's a PDB ID
    pdb_id = pdb_input.upper()
    
    # Fetch PDB content
    try:
        url = f"https://files.rcsb.org/download/{pdb_id}.pdb"
        response = requests.get(url, timeout=30)
        response.raise_for_status()
        pdb_content = response.text
        logger.info(f"Successfully fetched PDB {pdb_id} ({len(pdb_content)} characters)")
    except Exception as e:
        return {
            "success": False,
            "uuid": None,
            "message": f"Failed to fetch PDB {pdb_id}: {str(e)}",
            "pdb_id": pdb_id,
            "instructions": "Please verify the PDB ID is correct"
        }
    
    # Attempt to upload
    uuid, upload_status = attempt_pdb_upload(pdb_id, pdb_content, folder_uuid)
    
    if uuid:
        return {
            "success": True,
            "uuid": uuid,
            "message": upload_status,
            "pdb_id": pdb_id,
            "instructions": None
        }
    
    # Upload failed - provide instructions
    instructions = f"""
To use PDB {pdb_id} for docking:

1. Download the PDB file:
   https://files.rcsb.org/download/{pdb_id}.pdb

2. Upload to your Rowan account:
   https://labs.rowansci.com

3. Get the protein UUID from your account

4. Use the UUID in your docking call:
   rowan_docking(
       target_uuid="your-protein-uuid-here",
       ...
   )
"""
    
    return {
        "success": False,
        "uuid": None,
        "message": upload_status,
        "pdb_id": pdb_id,
        "instructions": instructions
    }

def enhance_docking_with_pdb_handler(
    target_input: Optional[str] = None,
    target_uuid: Optional[str] = None,
    **kwargs
) -> Tuple[Optional[str], str]:
    """
    Enhanced target handling for docking that accepts PDB IDs.
    
    Returns:
        Tuple of (target_uuid or None, message)
    """
    # If UUID already provided, use it
    if target_uuid:
        return target_uuid, "Using provided target UUID"
    
    # If target_input provided, try to handle it
    if target_input:
        result = handle_pdb_input(target_input, kwargs.get('folder_uuid'))
        
        if result['success']:
            return result['uuid'], result['message']
        else:
            # Return None but with helpful message
            return None, result['instructions'] or result['message']
    
    return None, "No protein target specified"


"""
Rowan v2 API: Redox Potential Workflow
Calculate reduction and oxidation potentials for molecules.
"""

from typing import Annotated
import rowan
import stjames


def submit_redox_potential_workflow(
    initial_molecule: Annotated[str, "SMILES string for redox potential calculation"],
    reduction: Annotated[bool, "Whether to calculate reduction potential (gaining electron)"] = False,
    oxidization: Annotated[bool, "Whether to calculate oxidation potential (losing electron)"] = True,
    mode: Annotated[str, "Calculation mode: 'rapid' (fast), 'careful' (balanced), or 'meticulous' (thorough)"] = "rapid",
    name: Annotated[str, "Workflow name for identification and tracking"] = "Redox Potential Workflow",
    folder_uuid: Annotated[str, "UUID of folder to organize this workflow. Empty string uses default folder"] = "",
    max_credits: Annotated[int, "Maximum credits to spend on this calculation. 0 for no limit"] = 0
):
    """Submit a redox potential calculation workflow using Rowan v2 API.
    
    Args:
        initial_molecule: SMILES string for redox potential calculation
        reduction: Whether to calculate reduction potential (gaining electron)
        oxidization: Whether to calculate oxidation potential (losing electron)
        mode: Calculation mode: 'rapid' (fast), 'careful' (balanced), or 'meticulous' (thorough)
        name: Workflow name for identification and tracking
        folder_uuid: UUID of folder to organize this workflow. Empty string uses default folder.
        max_credits: Maximum credits to spend on this calculation. 0 for no limit.
    
    Calculates reduction and/or oxidation potentials for a molecule using
    quantum chemistry methods.
    
    Returns:
        Workflow object representing the submitted workflow
        
    Example:
        # Simple redox potential from SMILES
        result = submit_redox_potential_workflow(
            initial_molecule="Cc1ccccc1",  # Toluene
            reduction=True,
            oxidization=True,
            name="Toluene Redox Potential"
        )
    """
    
    return rowan.submit_redox_potential_workflow(
        initial_molecule=stjames.Molecule.from_smiles(initial_molecule),
        reduction=reduction,
        oxidization=oxidization,
        mode=mode,
        name=name,
        folder_uuid=folder_uuid if folder_uuid else None,
        max_credits=max_credits if max_credits > 0 else None
    )


"""
Calculate the conformers of a molecule using the Rowan API.

| Mode                                          | Reckless | Rapid | Careful | Meticulous |
|-----------------------------------------------|----------|-------|---------|------------|
| Conformer generation program                  | RDKit    | RDKit | CREST   | CREST      |
| Level of theory used for conformer generation | MMFF     | MMFF  | GFN-FF  | GFN2-xTB   |
| CREST Mode                                    |          |       | quick   | normal     |
| ETKDG number of initial conformers            | 200      | 300   |         |            |
| Initial energy cutoff (kcal/mol)              | 10       | 15    | 10      | 15         |
| RMSD similarity cutoff (Å)                    | 0.25     | 0.10  | 0.00    | 0.00       |
| xTB screening max number of conformers        | 50       | 100   | 150     | 500        |
| Final energy cutoff (kcal/mol)                | 5        | 5     | 5       | 10         |

Rapid is recommended for most work.

See documentation at: https://docs.rowansci.com/science/workflows/conformers
"""

from stjames import Molecule

import rowan

# rowan.api_key = ""

# Run calculation remotely
result = rowan.submit_conformer_search_workflow(
    initial_molecule=Molecule.from_smiles("CCOCC"),
)

print(result)


"""
Rowan scan analysis functions for MCP tool integration.
"""

from typing import Optional, Dict, Any, Tuple, List
import rowan

def rowan_scan_analyzer(
    scan_uuid: str,
    action: str = "analyze",
    energy_threshold: Optional[float] = None
) -> str:
    """Analyze scan results and extract key geometries for IRC workflows.
    
    ** Essential IRC Tool:**
    - Analyzes completed scan workflows to extract transition state geometries
    - Provides formatted results ready for IRC calculations
    - Identifies energy maxima, minima, and barriers automatically
    
    ** Analysis Actions:**
    - **analyze**: Complete analysis with energy profile and key points (default)
    - **extract_ts**: Extract highest energy geometry (TS approximation for IRC)
    - **extract_minima**: Extract low energy geometries 
    - **energy_profile**: Show energy vs coordinate data for plotting
    
    ** IRC Workflow Integration:**
    1. Run scan → get scan_uuid
    2. Use: rowan_scan_analyzer(scan_uuid, "extract_ts")
    3. Copy TS geometry for transition state optimization
    4. Run IRC from optimized TS
    
    ** Example Usage:**
    - Full analysis: rowan_scan_analyzer("uuid-123", "analyze")
    - Extract TS: rowan_scan_analyzer("uuid-123", "extract_ts")
    - Find minima: rowan_scan_analyzer("uuid-123", "extract_minima", energy_threshold=2.0)
    
    Args:
        scan_uuid: UUID of the completed scan workflow to analyze
        action: Analysis type ("analyze", "extract_ts", "extract_minima", "energy_profile")
        energy_threshold: Energy threshold in kcal/mol above minimum for minima extraction (default: None)
    
    Returns:
        Analysis results with geometries, energies, and IRC preparation instructions
    """
    
    action = action.lower()
    
    try:
        # Retrieve scan workflow results
        workflow = rowan.Workflow.retrieve(uuid=scan_uuid)
        
        # Check if workflow is completed
        status = workflow.get('object_status', -1)
        if status != 2:  # Not completed
            status_names = {0: "Queued", 1: "Running", 3: "Failed", 4: "Stopped"}
            return f" Scan workflow is not completed. Status: {status_names.get(status, 'Unknown')}"
        
        # Extract scan data from object_data
        object_data = workflow.get('object_data', {})
        if not object_data:
            return " No scan data found in workflow results"
        
        # Parse scan results
        scan_results = parse_scan_data(object_data)
        if not scan_results:
            return " Could not parse scan data"
        
        # Perform requested analysis
        if action == "analyze":
            return format_full_analysis(scan_results, scan_uuid)
        elif action == "extract_ts":
            return extract_ts_geometry(scan_results, scan_uuid)
        elif action == "extract_minima":
            return extract_minima_geometries(scan_results, energy_threshold)
        elif action == "energy_profile":
            return format_energy_profile(scan_results)
        else:
            return f" Unknown action '{action}'. Available: analyze, extract_ts, extract_minima, energy_profile"
            
    except Exception as e:
        return f" Error analyzing scan: {str(e)}"

def parse_scan_data(object_data: Dict[str, Any]) -> Optional[Dict[str, Any]]:
    """Parse raw scan data into structured format."""
    
    try:
        # Look for scan points in various possible locations
        scan_points = None
        
        # Try different object_data structures
        if 'scan_points' in object_data:
            scan_points = object_data['scan_points']
        elif 'results' in object_data and isinstance(object_data['results'], list):
            scan_points = object_data['results']
        elif 'calculations' in object_data:
            scan_points = object_data['calculations']
        else:
            # Try to find array-like data
            for key, value in object_data.items():
                if isinstance(value, list) and len(value) > 1:
                    scan_points = value
                    break
        
        if not scan_points:
            return None
        
        # Structure the scan data
        parsed_data = {
            'points': [],
            'energies': [],
            'coordinates': [],
            'geometries': []
        }
        
        for i, point in enumerate(scan_points):
            if isinstance(point, dict):
                # Extract energy
                energy = point.get('energy') or point.get('total_energy') or point.get('scf_energy')
                if energy is not None:
                    parsed_data['energies'].append(energy)
                
                # Extract coordinate value
                coord_val = point.get('coordinate_value') or point.get('scan_coordinate') or i
                parsed_data['coordinates'].append(coord_val)
                
                # Extract geometry (XYZ coordinates)
                geometry = extract_geometry_from_point(point)
                parsed_data['geometries'].append(geometry)
                
                parsed_data['points'].append({
                    'index': i,
                    'energy': energy,
                    'coordinate': coord_val,
                    'geometry': geometry
                })
        
        return parsed_data if parsed_data['energies'] else None
        
    except Exception as e:
        print(f"Error parsing scan data: {e}")
        return None

def extract_geometry_from_point(point_data: Dict[str, Any]) -> Optional[str]:
    """Extract XYZ geometry from a scan point."""
    
    try:
        # Look for geometry in various formats
        if 'geometry' in point_data:
            geom = point_data['geometry']
            if isinstance(geom, str):
                return geom
            elif isinstance(geom, dict) and 'xyz' in geom:
                return geom['xyz']
        
        # Look for molecule object
        if 'molecule' in point_data:
            mol = point_data['molecule']
            if isinstance(mol, dict) and 'geometry' in mol:
                return mol['geometry']
        
        # Look for atoms/coordinates
        if 'atoms' in point_data:
            atoms = point_data['atoms']
            if isinstance(atoms, list):
                return format_atoms_to_xyz(atoms)
        
        return None
        
    except Exception:
        return None

def format_atoms_to_xyz(atoms: List[Dict]) -> str:
    """Convert atoms list to XYZ format."""
    
    try:
        xyz_lines = [str(len(atoms)), ""]  # Number of atoms + comment line
        
        for atom in atoms:
            symbol = atom.get('symbol') or atom.get('element')
            coords = atom.get('coordinates') or atom.get('position') or [0, 0, 0]
            
            if symbol and len(coords) >= 3:
                xyz_lines.append(f"{symbol:2s} {coords[0]:12.6f} {coords[1]:12.6f} {coords[2]:12.6f}")
        
        return '\n'.join(xyz_lines)
        
    except Exception:
        return ""

def extract_ts_geometry(scan_results: Dict[str, Any], scan_uuid: str) -> str:
    """Extract the highest energy geometry (TS approximation)."""
    
    energies = scan_results['energies']
    geometries = scan_results['geometries']
    coordinates = scan_results['coordinates']
    
    if not energies:
        return " No energy data found in scan results"
    
    # Find highest energy point
    max_energy_idx = energies.index(max(energies))
    max_energy = energies[max_energy_idx]
    max_coord = coordinates[max_energy_idx]
    ts_geometry = geometries[max_energy_idx]
    
    # Calculate relative energy (kcal/mol above minimum)
    min_energy = min(energies)
    rel_energy = (max_energy - min_energy) * 627.509  # Hartree to kcal/mol
    
    formatted = f" **Transition State Approximation Extracted**\n\n"
    formatted += f"**Scan Point:** {max_energy_idx + 1} of {len(energies)}\n"
    formatted += f" **Coordinate Value:** {max_coord:.3f}\n"
    formatted += f" **Energy:** {max_energy:.6f} hartree\n"
    formatted += f" **Barrier:** {rel_energy:.2f} kcal/mol above minimum\n\n"
    
    if ts_geometry:
        formatted += f" **XYZ Geometry:**\n```\n{ts_geometry}\n```\n\n"
        formatted += f" **Next Steps:**\n"
        formatted += f"1. Use this geometry for transition state optimization\n"
        formatted += f"2. Run: `rowan_spin_states(name='TS_opt', molecule='<geometry>', transition_state=True)`\n"
        formatted += f"3. Verify single imaginary frequency\n"
        formatted += f"4. Use optimized TS for IRC calculation\n"
    else:
        formatted += f" **Warning:** Could not extract geometry data\n"
        formatted += f" Check scan workflow {scan_uuid} manually for geometry data\n"
    
    return formatted

def extract_minima_geometries(scan_results: Dict[str, Any], energy_threshold: Optional[float] = None) -> str:
    """Extract low energy geometries (minima)."""
    
    energies = scan_results['energies']
    geometries = scan_results['geometries']
    coordinates = scan_results['coordinates']
    
    if not energies:
        return " No energy data found in scan results"
    
    min_energy = min(energies)
    threshold = energy_threshold or 2.0  # Default 2 kcal/mol above minimum
    threshold_hartree = threshold / 627.509
    
    # Find all points within energy threshold
    minima_points = []
    for i, energy in enumerate(energies):
        if energy <= min_energy + threshold_hartree:
            rel_energy = (energy - min_energy) * 627.509
            minima_points.append({
                'index': i,
                'energy': energy,
                'rel_energy': rel_energy,
                'coordinate': coordinates[i],
                'geometry': geometries[i]
            })
    
    formatted = f"**Low Energy Structures** (within {threshold:.1f} kcal/mol)\n\n"
    formatted += f"Found {len(minima_points)} structures:\n\n"
    
    for point in minima_points:
        formatted += f"**Point {point['index'] + 1}:**\n"
        formatted += f"   Coordinate: {point['coordinate']:.3f}\n"
        formatted += f"   Energy: +{point['rel_energy']:.2f} kcal/mol\n"
        if point['geometry']:
            # Show first few lines of geometry
            geom_lines = point['geometry'].split('\n')[:5]
            formatted += f"   Geometry: {' | '.join(geom_lines[:2])}\n"
        formatted += "\n"
    
    return formatted

def format_energy_profile(scan_results: Dict[str, Any]) -> str:
    """Format energy profile data for plotting/analysis."""
    
    energies = scan_results['energies']
    coordinates = scan_results['coordinates']
    
    if not energies:
        return " No energy data found"
    
    min_energy = min(energies)
    
    formatted = f" **Energy Profile Data**\n\n"
    formatted += f"Coordinate | Energy (hartree) | Rel Energy (kcal/mol)\n"
    formatted += f"-----------|------------------|--------------------\n"
    
    for coord, energy in zip(coordinates, energies):
        rel_energy = (energy - min_energy) * 627.509
        formatted += f"{coord:10.3f} | {energy:15.6f} | {rel_energy:18.2f}\n"
    
    formatted += f"\n **Energy Range:** {(max(energies) - min_energy) * 627.509:.2f} kcal/mol\n"
    formatted += f" **Barrier Location:** Coordinate {coordinates[energies.index(max(energies))]:.3f}\n"
    
    return formatted

def format_full_analysis(scan_results: Dict[str, Any], scan_uuid: str) -> str:
    """Provide complete scan analysis."""
    
    energies = scan_results['energies']
    coordinates = scan_results['coordinates']
    
    if not energies:
        return " No scan data to analyze"
    
    min_energy = min(energies)
    max_energy = max(energies)
    energy_range = (max_energy - min_energy) * 627.509
    
    # Find key points
    min_idx = energies.index(min_energy)
    max_idx = energies.index(max_energy)
    
    formatted = f" **Complete Scan Analysis**\n\n"
    formatted += f" **Overview:**\n"
    formatted += f"  • Total Points: {len(energies)}\n"
    formatted += f"  • Energy Range: {energy_range:.2f} kcal/mol\n"
    formatted += f"  • Coordinate Range: {min(coordinates):.3f} to {max(coordinates):.3f}\n\n"
    
    formatted += f"🌊 **Global Minimum:**\n"
    formatted += f"  • Point {min_idx + 1}: {coordinates[min_idx]:.3f}\n"
    formatted += f"  • Energy: {min_energy:.6f} hartree\n\n"
    
    formatted += f"⛰ **Energy Maximum (TS Approx):**\n"
    formatted += f"  • Point {max_idx + 1}: {coordinates[max_idx]:.3f}\n"
    formatted += f"  • Energy: {max_energy:.6f} hartree\n"
    formatted += f"  • Barrier: {energy_range:.2f} kcal/mol\n\n"
    
    formatted += f" **Recommended Next Steps:**\n"
    formatted += f"1. Extract TS geometry: `rowan_scan_analyzer('{scan_uuid}', 'extract_ts')`\n"
    formatted += f"2. Optimize transition state with extracted geometry\n"
    formatted += f"3. Run IRC from optimized TS\n"
    
    return formatted

def test_rowan_scan_analyzer():
    """Test the scan analyzer function."""
    try:
        # Test with dummy UUID 
        result = rowan_scan_analyzer("test-scan-uuid", "analyze")
        print(" Scan analyzer test completed!")
        print(f"Result type: {type(result)}")
        return True
    except Exception as e:
        print(f" Scan analyzer test failed: {e}")
        return False

if __name__ == "__main__":
    test_rowan_scan_analyzer() 


"""
Rowan v2 API: Protein Cofolding Workflow
Simulate protein-protein interactions and cofolding.
"""

from typing import List, Annotated
import rowan
import stjames
import json


def submit_protein_cofolding_workflow(
    initial_protein_sequences: Annotated[str, "JSON string list of protein sequences for cofolding (e.g., '[\"MKLLV...\", \"MAHQR...\"]')"],
    initial_smiles_list: Annotated[str, "JSON string list of SMILES for ligands to include in cofolding (e.g., '[\"CCO\", \"CC(=O)O\"]'). Empty for protein-only"] = "",
    ligand_binding_affinity_index: Annotated[str, "JSON string mapping ligand indices to protein binding sites. Empty for automatic detection"] = "",
    use_msa_server: Annotated[bool, "Whether to use multiple sequence alignment server for better structure prediction"] = True,
    use_potentials: Annotated[bool, "Whether to include additional potentials in the calculation"] = False,
    name: Annotated[str, "Workflow name for identification and tracking"] = "Cofolding Workflow",
    model: Annotated[str, "Structure prediction model to use (e.g., 'boltz_2', 'alphafold3')"] = "boltz_2",
    folder_uuid: Annotated[str, "UUID of folder to organize this workflow. Empty string uses default folder"] = "",
    max_credits: Annotated[int, "Maximum credits to spend on this calculation. 0 for no limit"] = 0
):
    """Submits a protein cofolding workflow to the API.
    
    Args:
        initial_protein_sequences: JSON string list of protein sequences (amino acid strings) to cofold
        initial_smiles_list: JSON string list of ligand SMILES strings to include in cofolding. Empty string for protein-only
        ligand_binding_affinity_index: Index of ligand in initial_smiles_list for binding affinity calculation (e.g., "0"). Empty string skips affinity
        use_msa_server: Whether to use MSA (Multiple Sequence Alignment) server for improved accuracy
        use_potentials: Whether to use statistical potentials in the calculation
        name: Workflow name for identification and tracking
        model: Cofolding model to use (e.g., 'boltz_2', 'alphafold3')
        folder_uuid: UUID of folder to organize this workflow. Empty string uses default folder.
        max_credits: Maximum credits to spend on this calculation. 0 for no limit.
    
    Returns:
        Workflow object representing the submitted workflow
        
    Example:
        # Protein-ligand cofolding with CDK2 kinase (from test)
        result = submit_protein_cofolding_workflow(
            initial_protein_sequences='["MENFQKVEKIGEGTYGVVYKARNKLTGEVVALKKIRLDTETEGVPSTAIREISLLKELNHPNIVKLLDVIHTENKLYLVFEFLHQDLKKFMDASALTGIPLPLIKSYLFQLLQGLAFCHSHRVLHRDLKPQNLLINTEGAIKLADFGLARAFGVPVRTYTHEVVTLWYRAPEILLGCKYYSTAVDIWSLGCIFAEMVTRRALFPGDSEIDQLFRIFRTLGTPDEVVWPGVTSMPDYKPSFPKWARQDFSKVVPPLDEDGRSLLSQMLHYDPNKRISAKAALAHPFFQDVTKPVPHLRL"]',
            initial_smiles_list='["CCC(C)(C)NC1=NCC2(CCC(=O)C2C)N1"]',
            ligand_binding_affinity_index="0",
            name="Cofolding CDK2 with ligand"
        )
    """
    # Parse initial_protein_sequences (always a string in simplified version)
    try:
        initial_protein_sequences = json.loads(initial_protein_sequences)
    except (json.JSONDecodeError, ValueError):
        # Try to parse as comma-separated
        if ',' in initial_protein_sequences:
            initial_protein_sequences = [s.strip() for s in initial_protein_sequences.split(',') if s.strip()]
        else:
            initial_protein_sequences = [initial_protein_sequences.strip()]
    
    # Parse initial_smiles_list if provided
    parsed_initial_smiles_list = None
    if initial_smiles_list:
        try:
            parsed_initial_smiles_list = json.loads(initial_smiles_list)
        except (json.JSONDecodeError, ValueError):
            # Try to parse as comma-separated
            if ',' in initial_smiles_list:
                parsed_initial_smiles_list = [s.strip() for s in initial_smiles_list.split(',') if s.strip()]
            else:
                parsed_initial_smiles_list = [initial_smiles_list.strip()]
    
    # Parse ligand_binding_affinity_index if provided
    parsed_ligand_index = None
    if ligand_binding_affinity_index:
        try:
            parsed_ligand_index = int(ligand_binding_affinity_index)
        except (ValueError, TypeError):
            raise ValueError(f"Invalid ligand_binding_affinity_index: '{ligand_binding_affinity_index}' must be an integer")
    
    return rowan.submit_protein_cofolding_workflow(
        initial_protein_sequences=initial_protein_sequences,
        initial_smiles_list=parsed_initial_smiles_list,
        ligand_binding_affinity_index=parsed_ligand_index,
        use_msa_server=use_msa_server,
        use_potentials=use_potentials,
        name=name,
        model=model,
        folder_uuid=folder_uuid if folder_uuid else None,
        max_credits=max_credits if max_credits > 0 else None
    )


"""
Rowan v2 API: Solubility Workflow
Predict molecular solubility in various solvents at different temperatures.
"""

from typing import List, Annotated
import rowan
import json


def submit_solubility_workflow(
    initial_smiles: Annotated[str, "SMILES string of the molecule for solubility prediction"],
    solvents: Annotated[str, "JSON string list of solvents as SMILES or names (e.g., '[\"water\", \"ethanol\", \"CCO\"]'). Empty string uses defaults"] = "",
    temperatures: Annotated[str, "JSON string list of temperatures in Kelvin (e.g., '[298.15, 310.15]'). Empty string uses default range"] = "",
    name: Annotated[str, "Workflow name for identification and tracking"] = "Solubility Workflow",
    folder_uuid: Annotated[str, "UUID of folder to organize this workflow. Empty string uses default folder"] = "",
    max_credits: Annotated[int, "Maximum credits to spend on this calculation. 0 for no limit"] = 0
):
    """Submit a solubility prediction workflow using Rowan v2 API.
    
    Args:
        initial_smiles: SMILES string of the molecule for solubility prediction
        solvents: JSON string list of solvents as SMILES or names (e.g., '["water", "ethanol", "CCO"]'). Empty string uses defaults
        temperatures: JSON string list of temperatures in Kelvin (e.g., '[298.15, 310.15]'). Empty string uses default range
        name: Workflow name for identification and tracking
        folder_uuid: UUID of folder to organize this workflow. Empty string uses default folder.
        max_credits: Maximum credits to spend on this calculation. 0 for no limit.
    
    Predicts solubility (log S) of a molecule in multiple solvents at various temperatures
    using machine learning models.
    
    Returns:
        Workflow object representing the submitted workflow
        
    Example:
        # Basic solubility prediction
        result = submit_solubility_workflow(
            initial_smiles="CC(=O)Nc1ccc(O)cc1",
            solvents='["water", "ethanol"]',
            temperatures='[298.15, 310.15]'
        )
        
        # With SMILES solvents
        result = submit_solubility_workflow(
            initial_smiles="CC(=O)O",
            solvents='["O", "CCO", "CCCCCC"]',
            temperatures='[273.15, 298.15, 323.15]'
        )
    """
    
    # Parse solvents parameter - handle string input
    parsed_solvents = None
    if solvents:
        # Handle various string formats
        solvents = solvents.strip()
        if solvents.startswith('[') and solvents.endswith(']'):
            # JSON array format like '["water", "ethanol"]'
            try:
                parsed_solvents = json.loads(solvents)
            except (json.JSONDecodeError, ValueError):
                # Failed to parse as JSON, try as comma-separated
                solvents = solvents.strip('[]').replace('"', '').replace("'", "")
                parsed_solvents = [s.strip() for s in solvents.split(',') if s.strip()]
        elif ',' in solvents:
            # Comma-separated format like 'water, ethanol'
            parsed_solvents = [s.strip() for s in solvents.split(',') if s.strip()]
        else:
            # Single solvent as string like 'water'
            parsed_solvents = [solvents]
        
        # Convert solvent names to SMILES if needed
        solvent_name_to_smiles = {
            'water': 'O',
            'ethanol': 'CCO',
            'dmso': 'CS(=O)C',
            'acetone': 'CC(=O)C',
            'methanol': 'CO',
            'chloroform': 'C(Cl)(Cl)Cl',
            'dichloromethane': 'C(Cl)Cl',
            'toluene': 'Cc1ccccc1',
            'benzene': 'c1ccccc1',
            'hexane': 'CCCCCC',
            'ether': 'CCOCC',
            'diethyl ether': 'CCOCC',
            'thf': 'C1CCOC1',
            'tetrahydrofuran': 'C1CCOC1',
            'dioxane': 'C1COCCO1',
            'acetonitrile': 'CC#N',
            'pyridine': 'c1ccncc1'
        }
        
        converted_solvents = []
        for solvent in parsed_solvents:
            solvent_lower = solvent.lower().strip()
            if solvent_lower in solvent_name_to_smiles:
                converted_solvents.append(solvent_name_to_smiles[solvent_lower])
            else:
                # Assume it's already a SMILES string or use as-is
                converted_solvents.append(solvent)
        parsed_solvents = converted_solvents
        
        # Validate the final solvent SMILES to catch issues early
        try:
            from rdkit import Chem
            for i, solvent_smiles in enumerate(parsed_solvents):
                mol = Chem.MolFromSmiles(solvent_smiles)
                if mol is None:
                    raise ValueError(f"Invalid solvent SMILES: '{solvent_smiles}' at position {i}")
        except ImportError:
            # RDKit not available for validation, proceed anyway
            pass
    
    # Parse temperatures parameter - handle string input
    parsed_temperatures = None
    if temperatures:
        # Handle various string formats
        temperatures = temperatures.strip()
        if temperatures.startswith('[') and temperatures.endswith(']'):
            # JSON array format like '[298.15, 310.15]'
            try:
                parsed_temperatures = json.loads(temperatures)
            except (json.JSONDecodeError, ValueError):
                # Failed to parse as JSON, try as comma-separated
                temperatures = temperatures.strip('[]').replace('"', '').replace("'", "")
                parsed_temperatures = [float(t.strip()) for t in temperatures.split(',') if t.strip()]
        elif ',' in temperatures:
            # Comma-separated format like '298.15, 310.15'
            parsed_temperatures = [float(t.strip()) for t in temperatures.split(',') if t.strip()]
        else:
            # Single temperature as string like '298.15'
            parsed_temperatures = [float(temperatures)]
    
    # Validate the main SMILES string early to catch issues
    try:
        from rdkit import Chem
        mol = Chem.MolFromSmiles(initial_smiles)
        if mol is None:
            raise ValueError(f"Invalid initial SMILES: '{initial_smiles}'")
    except ImportError:
        # RDKit not available for validation, proceed anyway
        pass
    
    try:
        result = rowan.submit_solubility_workflow(
            initial_smiles=initial_smiles,
            solvents=parsed_solvents,
            temperatures=parsed_temperatures,
            name=name,
            folder_uuid=folder_uuid if folder_uuid else None,
            max_credits=max_credits if max_credits > 0 else None
        )
        
        return result
        
    except Exception as e:
        # Re-raise the exception so MCP can handle it
        raise


import stjames

import rowan

phenols_to_compute = {
    "p-nitrophenol": "Oc1ccc(N(=O)=O)cc1",
    "p-(trifluoromethyl)phenol": "Oc1ccc(C(F)(F)F)cc1",
    "methyl p-hydroxybenzoate": "COC(=O)c1ccc(O)cc1",
    "p-fluorophenol": "Oc1ccc(F)cc1",
    "p-chlorophenol": "Oc1ccc(Cl)cc1",
    "phenol": "Oc1ccccc1",
    "p-cresol": "Oc1ccc(C)cc1",
    "p-methoxyphenol": "COc1ccc(O)cc1",
    "p-(dimethylamino)phenol": "CN(C)c1ccc(O)cc1",
}

pka_folder = rowan.create_folder(name="phenol pKa results")

workflows = []
for name, smiles in phenols_to_compute.items():
    stjames_molecule = stjames.Molecule.from_smiles(smiles)
    workflows.append(
        rowan.submit_pka_workflow(
            stjames_molecule,
            name=f"pKa {name}",
            folder_uuid=pka_folder.uuid,
            deprotonate_elements=[8],
        )
    )


for workflow in workflows:
    workflow.wait_for_result()
    workflow.fetch_latest(in_place=True)

print([(workflow.name, workflow.data["conjugate_bases"][0]["pka"]) for workflow in workflows])


"""
Folder management operations for Rowan API.
"""

import os
import rowan
from typing import Optional

# Set up logging
import logging
logger = logging.getLogger(__name__)



def rowan_folder_management(
    action: str,
    folder_uuid: Optional[str] = None,
    name: Optional[str] = None,
    parent_uuid: Optional[str] = None,
    notes: Optional[str] = None,
    starred: Optional[bool] = None,
    public: Optional[bool] = None,
    created_at: Optional[str] = None,
    updated_at: Optional[str] = None,
    is_root: Optional[bool] = None,
    uuid: Optional[str] = None,
    name_contains: Optional[str] = None,
    page: int = 0,
    size: int = 50
) -> str:
    """Unified folder management tool for all folder operations. Available actions: create, retrieve, update, delete, list.
    
    **Available Actions:**
    - **create**: Create a new folder (requires: name, optional: parent_uuid, notes, starred, public)
    - **retrieve**: Get folder details (requires: folder_uuid)
    - **update**: Update folder properties (requires: folder_uuid, optional: name, parent_uuid, notes, starred, public)
    - **delete**: Delete a folder (requires: folder_uuid)
    - **list**: List folders with filters (optional: name_contains, parent_uuid, starred, public, page, size)
    
    Args:
        action: Action to perform ('create', 'retrieve', 'update', 'delete', 'list')
        folder_uuid: UUID of the folder (required for retrieve, update, delete)
        name: Folder name (required for create, optional for update)
        parent_uuid: Parent folder UUID (optional for create/update, if not provided creates in root)
        notes: Folder notes (optional for create/update)
        starred: Star the folder (optional for create/update)
        public: Make folder public (optional for create/update)
        created_at: The date and time at which this folder was created
        updated_at: The date and time at which this folder was most recently updated
        is_root: Whether or not this folder is the user's root folder
        uuid: The UUID of this folder
        name_contains: Filter by name containing text (optional for list)
        page: Page number for pagination (default: 1, for list)
        size: Results per page (default: 50, for list)
    
    Returns:
        Results of the folder operation
    """
    
    action = action.lower()
    
    try:
        if action == "create":
            if not name:
                return "Error: 'name' is required for creating a folder"
            
            return rowan.Folder.create(
                name=name,
                parent_uuid=parent_uuid,
                notes=notes or "",
                starred=starred or False,
                public=public or False
            )
            
        elif action == "retrieve":
            if not folder_uuid:
                return "Error: 'folder_uuid' is required for retrieving a folder"
            
            return rowan.Folder.retrieve(uuid=folder_uuid)
            
        elif action == "update":
            if not folder_uuid:
                return "Error: 'folder_uuid' is required for updating a folder"

            return rowan.Folder.update(
                uuid=folder_uuid,
                name=name,
                notes=notes,
                starred=starred,
                public=public,
                parent_uuid=parent_uuid
                
            )
            
        elif action == "delete":
            if not folder_uuid:
                return "Error: 'folder_uuid' is required for deleting a folder"
            
            rowan.Folder.delete(uuid=folder_uuid)
            return "Folder deleted successfully"
            
        elif action == "list":
            return rowan.Folder.list(
                name_contains=name_contains,
                parent_uuid=parent_uuid,
                starred=starred,
                public=public,
                page=page,
                size=size
            )
              
    except Exception as e:
        return f"Error in folder {action}: {str(e)}"


def test_rowan_folder_management():
    """Test the rowan_folder_management function."""
    try:
        # Test listing folders
        result = rowan_folder_management(action="list", size=5)
        print("Folder management test successful!")
        print(f"Result: {result[:200]}...")  # Show first 200 chars
        return True
    except Exception as e:
        print(f"Folder management test failed: {e}")
        return False


if __name__ == "__main__":
    test_rowan_folder_management() 